import os.path, pathlib
import sys
import weakref, io, asyncio, threading, time
from .HTML import CSS, HTML, HTMLManager
from .WidgetTools import JupyterAPIs, DefaultOutputWidget

__all__ = [
    "HTMLWidgets",
    "ActiveHTMLWrapper"
]

__reload_hook__ = [".HTML", ".WidgetTools"]

class ActiveHTMLWrapper:
    base=None
    _subwrappers = weakref.WeakKeyDictionary()
    # to keep anonymous wrappers alive until their elements go out of scope
    def __init__(self,
                 *elements,
                 tag=None,
                 cls=None,
                 id=None,
                 value=None,
                 style=None,
                 event_handlers=None,
                 javascript_handles=None,
                 onevents=None,
                 data=None,
                 debug_pane=None,
                 track_value=None,
                 continuous_update=None,
                 **attributes
                 ):
        HTMLElement = self.load_HTMLElement()
        self.HTMLElement = HTMLElement

        if len(elements) == 1 and isinstance(elements[0], (list, tuple)):
            elements = elements[0]

        #pre-prune
        attrs = {}
        if '_debugPrint' in attributes:
            attrs['_debugPrint'] = attributes['_debugPrint']
            del attributes['_debugPrint']
        if 'trackInput' in attributes:
            if track_value is None:
                track_value = attributes['trackInput']
            del attributes['trackInput']
        if track_value is not None:
            attrs['trackInput'] = track_value
        if 'continuousUpdate' in attributes:
            if continuous_update is None:
                continuous_update = attributes['continuousUpdate']
            del attributes['continuousUpdate']
        if continuous_update is not None:
            attrs['continuousUpdate'] = continuous_update

        self._handle_api = None
        if javascript_handles is not None:
            if hasattr(javascript_handles, 'javascript_handles'):
                self._handle_api = javascript_handles
                attrs['jsAPI'] = javascript_handles.elem
            else:
                if javascript_handles is not None and 'api' in javascript_handles:
                    javascript_handles = javascript_handles.copy()
                    self._handle_api = javascript_handles['api']
                    attrs['jsAPI'] = javascript_handles['api'].elem
                    del javascript_handles['api']
                attrs['jsHandlers'] = javascript_handles

        if self.base is not None:
            shadow_el = self.base(cls=cls, style=style, **attributes)
        else:
            shadow_el = None

        attributes = HTMLManager.manage_attrs(attributes)
        if tag is None and hasattr(shadow_el, 'tag'):
            tag = shadow_el.tag
        if tag is not None:
            attrs['tagName'] = tag

        if shadow_el is not None:
            if 'class' in shadow_el.attrs:
                cls = shadow_el['class']
        if cls is not None:
            classList = HTMLManager.manage_class(cls)
            attrs['classList'] = classList

        eventPropertiesDict = {}
        if event_handlers is not None:
            event_handlers = event_handlers.copy()
            for c, v in event_handlers.items():
                if isinstance(v, dict):
                    v = v.copy()
                    if 'callback' in v:
                        callback = v['callback']
                        del v['callback']
                        event_handlers[c] = callback
                        v['notify'] = True
                    else:
                        event_handlers[c] = None
                    eventPropertiesDict[c] = v
                elif isinstance(v, str):
                    eventPropertiesDict[c] = v
                    event_handlers[c] = None
                else:
                    eventPropertiesDict[c] = None
                    event_handlers[c] = v
        if len(eventPropertiesDict) > 0:
            attrs['eventPropertiesDict'] = eventPropertiesDict

        if onevents is not None:
            if event_handlers is None:
                event_handlers = {}
            for c,v in onevents.items():
                if isinstance(v, dict):
                    v = v.copy()
                    if 'callback' in v:
                        callback = v['callback']
                        del v['callback']
                        event_handlers[c] = callback
                        v['notify'] = True
                    else:
                        event_handlers[c] = None
                    onevents[c] = v
                elif isinstance(v, str):
                    onevents[c] = v
                    event_handlers[c] = None
                else:
                    onevents[c] = None
                    event_handlers[c] = v
            if len(onevents) > 0:
                attrs['onevents'] = onevents

        if not isinstance(elements, (list, tuple)):
            val = elements
            if isinstance(val, str):
                attrs['innerHTML'] = val
            elif isinstance(val, HTML.XMLElement):
                val.add_update_callback(
                    lambda w,*_,s=self:setattr(s, 'html', w) if hasattr(s, 'elem') else None
                )
                attrs['innerHTML'] = val.tostring()
            else:
                attrs['children'] = [self.canonicalize_widget(val)]
        elif len(elements) == 1:
            val = elements[0]
            if isinstance(val, str):
                attrs['innerHTML'] = val
            elif isinstance(val, HTML.XMLElement):
                val.add_update_callback(
                    lambda w,*_,s=self:setattr(s, 'html', w) if hasattr(s, 'elem') else None
                )
                attrs['innerHTML'] = val.tostring()
            else:
                attrs['children'] = [self.canonicalize_widget(val)]
        else:
            elements = [self.canonicalize_widget(x) for x in elements]
            attrs['children'] = elements
        if id is not None:
            attrs['id'] = id
        if value is not None:
            attrs['value'] = value

        extra_styles, attributes = HTMLManager.extract_styles(attributes)
        if shadow_el is not None:
            if 'style' in shadow_el.attrs:
                extra_styles = shadow_el['style']
            else:
                extra_styles = {}

        if style is None:
            style = extra_styles
        else:
            for k,v in extra_styles.items():
                if k in style and style[k] != v:
                    raise ValueError("style {} given twice".format(k))
                style[k] = v
        if len(style) > 0:
            style = HTMLManager.manage_styles(style).props
            # if isinstance(style, str):
            #     style = CSS.parse(style).props
            # elif isinstance(style, CSS):
            #     style = style.props
            attrs['styleDict'] = style

        if len(attributes) > 0:
            attrs['elementAttributes'] = attributes

        if data is not None:
            attrs['exportData'] = data

        self.elem = HTMLElement(**attrs)
        self.link(self.elem)
        self._html_cache = None

        if debug_pane is None:
            debug_pane = DefaultOutputWidget.get_default()
        self.debug_pane = debug_pane

        self._event_handlers = {} if event_handlers is None else event_handlers
        if len(eventPropertiesDict) > 0:
            self.elem.bind_callback(self.handle_event)
        # if eventPropertiesDict
        # for e in event_handlers:
        #     eventPropertiesDict[e] = None
    def __call__(self, *elems, **kwargs):
        return type(self)(
            self.elements + list(*elems),
            tag=self.tag,
            cls=self.class_list,
            style=self.style,
            event_handlers=self.event_handlers,
            onevents=self.onevents,
            javascript_handles=self.javascript_handles,
            value=self.value,
            id=self.id,
            debug_pane=self.debug_pane,
            track_value=self.track_value,
            continuous_update=self.continuous_update,
            **dict(self.attrs, **kwargs)
        )
    @classmethod
    def canonicalize_widget(cls, x):
        if isinstance(x, HTML.XMLElement.atomic_types):
            x = str(x)
        Widget = JupyterAPIs.get_widgets_api().Widget
        if isinstance(x, Widget):
            return x
        elif isinstance(x, ActiveHTMLWrapper):
            elem = x.elem
            cls._subwrappers[elem] = x
            return elem
        elif isinstance(x, str):
            return cls.canonicalize_widget(HTML.parse(x, strict=False))
        elif isinstance(x, HTML.XMLElement):
            subwrapper = ActiveHTMLWrapper.from_HTML(x)
            elem = subwrapper.elem
            cls._subwrappers[elem] = subwrapper
            return elem
        elif hasattr(x, 'to_widget'):
            return cls.canonicalize_widget(x.to_widget())
        else:
            raise NotImplementedError("don't know what to do with object {} of type {}".format(
                x, type(x)
            ))
    @classmethod
    def from_HTML(cls, x:HTML.XMLElement, event_handlers=None, debug_pane=None, **props):
        attrs = x.attrs
        props["event_handlers"] = event_handlers
        props["debug_pane"] = debug_pane
        for key, target in [("cls", "class")]:
            if target in attrs:
                props[key] = attrs[target]
                del attrs[target]
        props = dict(props, **attrs)
        body = []
        for y in x.elems:
            if hasattr(y, 'tostring'):
                # if hasattr(y, 'to_widget'):
                #     raise ValueError(y)
                # try:
                y = y.tostring()
                # except:
                #     raise Exception(y)
            body.append(y)

        new = cls(*body, tag=x.tag, **props)
        return new

    @classmethod
    def load_HTMLElement(cls):
        from .ActiveHTMLWidget import HTMLElement
        return HTMLElement
    @classmethod
    def convert_child(cls, c):
        HTMLElement = cls.load_HTMLElement()
        if isinstance(c, ActiveHTMLWrapper):
            return c.to_html()
        elif isinstance(c, HTMLElement):
            if hasattr(c, 'wrappers'):
                w = next(iter(c.wrappers))
                return w.to_html()
            else:
                try:
                    wrapper_cls = HTMLWidgets.get_class_map()[c.tagName]
                except KeyError:
                    wrapper_cls = lambda *elems, tag=c.tagName, **attrs: cls(*elems, tag=tag, **attrs)

                body = []
                if c.innerHTML != "":
                    body.append(c.innerHTML)
                elif c.textContent != "":
                    body.append(c.textContent)
                else:
                    body = [cls.convert_child(x) for x in c.children]
                attrs = c.elementAttributes.copy()
                if c.value != "":
                    attrs['value'] = c.value
                if c.id != "":
                    attrs["id"] = c.id
                if len(c.classList) > 0:
                    attrs["cls"] = c.classList
                if len(c.styleDict) > 0:
                    attrs["style"] = c.styleDict

                new_el = wrapper_cls(*body, **attrs)
                return new_el.to_html()
        elif isinstance(c, JupyterAPIs.get_widgets_api().Widget):
            html = HTML.XMLElement("jwidget", id=c.model_id)
            html.mirror = c
            return html
        elif isinstance(c, HTML.XMLElement):
            return c
        else:
            raise ValueError("don't know how to convert {} {}".format(type(c), HTMLElement))
    def to_html(self):
        try:
            wrapper_cls = HTML.get_class_map()[self.tag]
        except KeyError:
            wrapper_cls = lambda *elems, **attrs: HTML.XMLElement(self.tag, *elems, **attrs)

        # raise Exception(wrapper_cls, self.tag)
        attrs = self.attrs.copy()
        if self.value is not None:
            attrs['value'] = self.value
        if self.id is not None:
            attrs["id"] = self.id
        if len(self.class_list) > 0:
            attrs["class"] = self.class_list
        if len(self.style) > 0:
            attrs["style"] = self.style

        if self.html is not None:
            html = wrapper_cls(self.html, **attrs)
        elif self.text is not None:
            html = wrapper_cls(self.text, **attrs)
        else:
            html = wrapper_cls(
                *(self.convert_child(x) for x in self.children),
                **attrs
            )
        html.mirror = self
        return html

    def find(self, path, find_mirror=True):
        base = self.to_html().find(path)
        if find_mirror and base is not None:
            base = base.mirror
        return base
    def findall(self, path, find_mirror=True):
        base = self.to_html().findall(path)
        if find_mirror and base is not None:
            base = [b.mirror for b in base]
        return base
    def iterfind(self, path, find_mirror=True):
        bases = self.base.iterfind(path)
        for b in bases:
            if find_mirror and b is not None:
                b = b.mirror
            yield b
    def find_by_id(self, id, mode='first', parent=None, find_mirror=True):
        base = self.to_html().find_by_id(id, mode=mode, parent=parent)
        if find_mirror and base is not None:
            base = base.mirror
        return base

    # def to_tree(self, root=None, parent=None):
    #     if parent is not None:
    #         self._parents.add(parent)
    #     if self._tree_cache is None:
    #         ...
    #         if root is None:
    #             root = ElementTree.Element('root')
    #         attrs = self.construct_etree_attrs(self.attrs)
    #         my_el = ElementTree.SubElement(root, self.tag, attrs)
    #         if all(isinstance(e, str) for e in self.elems):
    #             my_el.text = "\n".join(self.elems)
    #         else:
    #             for elem in self.elems:
    #                 self.construct_etree_element(elem, my_el, parent=self)
    #         self._tree_cache = my_el
    #     elif root is not None:
    #         if self._tree_cache not in root:
    #             root.append(self._tree_cache)
    #     return self._tree_cache
    def to_widget(self, parent=None):
        return self.elem
    def __repr__(self):
        body = (
            self.children if len(self.children) > 0
            else self.html_string if self.html_string is not None
            else self.text if self.text is not None
            else ""
        )
        return "{}({}, {!r}, cls={}, style={}, id={})".format(
            type(self).__name__,
            self.tag,
            body,
            self.class_list,
            self.style,
            self.id
        )
    def display(self):
        wrapper = HTMLWidgets.Div(self, cls='jhtml')
        return JupyterAPIs.get_display_api().display(wrapper.elem)
    def _ipython_display_(self):
        return self.display()
    def get_mime_bundle(self):
        w = HTMLWidgets.Div(self, cls='jhtml').elem
        plaintext = repr(self)
        if len(plaintext) > 110:
            plaintext = plaintext[:110] + '…'
        data = {
            'text/plain': plaintext,
        }
        if w._view_name is not None:
            # The 'application/vnd.jupyter.widget-view+json' mimetype has not been registered yet.
            # See the registration process and naming convention at
            # http://tools.ietf.org/html/rfc6838
            # and the currently registered mimetypes at
            # http://www.iana.org/assignments/media-types/media-types.xhtml.
            data['application/vnd.jupyter.widget-view+json'] = {
                'version_major': 2,
                'version_minor': 0,
                'model_id': w._model_id
            }
        return data

    @staticmethod
    def _handle_event(e, event_handlers, self):
        try:
            handler = event_handlers[e['type']]
        except KeyError:
            pass
        else:
            handler(e, self)
    def handle_event(self, e):
        with self.debug_pane:
            return self._handle_event(e, self.event_handlers, self)

    def link(self, elem):
        if not hasattr(elem, 'wrappers'):
            elem.wrappers = weakref.WeakSet()
        elem.wrappers.add(self)
        def wrapper():
            return next(iter(elem.wrappers))
        elem.wrapper = wrapper

    # add an API to make the elem look more like regular HTML
    @property
    def tag(self):
        return 'div' if self.elem.tagName == '' else self.elem.tagName
    @property
    def id(self):
        eid = self.elem.id
        return None if eid == "" else eid
    @id.setter
    def id(self, val):
        if val is None:
            val = ""
        self.elem.id = val
    @property
    def text(self):
        eid = self.elem.textContent
        return None if eid == "" else eid
    @text.setter
    def text(self, val):
        self.elem.children = []
        self.html_string = ""
        self.elem.textContent = val
        self.elem.send_state('children')
    @property
    def value(self):
        eid = self.elem.value
        return None if eid == "" else eid
    @value.setter
    def value(self, val):
        if val is None:
            val = ""
        self.elem.value = val
    @property
    def attrs(self):
        return self.elem.elementAttributes
    @attrs.setter
    def attrs(self, val):
        if "id" in val:
            self.elem.id = val['id']
            del val['id']
        if "value" in val:
            self.elem.value = val['value']
            del val['value']
        if "javascript_handles" in val:
            self.javascript_handles = val
            del val['javascript_handles']
        val = HTMLManager.manage_attrs(val)
        if "class" in val:
            self.elem.classList = HTMLManager.manage_class(val['class'])
            self.elem.send_state('classList')
            del val['class']
        if "style" in val:
            self.elem.styleDict = HTMLManager.manage_styles(val["style"]).props
            del val['style']
            self.elem.send_state('styleDict')
        self.elem.elementAttributes = val

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.get_attribute(item)
        else:
            return self.get_child(item)
    def __setitem__(self, item, value):
        if isinstance(item, str):
            self.set_attribute(item, value)
        else:
            self.set_child(item, value)
    def __delitem__(self, item):
        if isinstance(item, str):
            self.del_attribute(item)
        else:
            self.del_child(item)
    def get_attribute(self, key):
        if key == "id":
            return self.id
        elif key == "value":
            return self.value
        elif key == "style":
            return self.style
        elif key == "javascript_handles":
            return self.javascript_handles
        elif key == "event_handlers":
            return self.event_handlers
        elif key == "onevents":
            return self.onevents
        else:
            key = HTMLManager.clean_key(key)
            if key == "class":
                return self.class_list
            if key in CSS.known_properties:
                return self.elem.styleDict[key]
            else:
                return self.elem.elementAttributes[key]
    def set_attribute(self, key, value):
        if key == "id":
            self.id = value
        elif key == "value":
            self.value = value
        elif key == "javascript_handles":
            self.javascript_handles = value
        elif key == "event_handlers":
            self.event_handlers = value
        elif key == "onevents":
            self.onevents = value
        elif key == "style":
            self.elem.styleDict = HTMLManager.manage_styles(value)
            self.elem.send_state('styleDict')
        else:
            key = HTMLManager.clean_key(key)
            if key == "class":
                self.class_list = value
            elif key in CSS.known_properties:
                self.elem.styleDict[key] = value
                self.elem.send_state('styleDict')
            else:
                attrs = self.elem.elementAttributes
                attrs[key] = value
                self.elem.elementAttributes = attrs
                self.elem.send_state('elementAttributes')
    def del_attribute(self, key):
        if key == "id":
            self.id = None
        elif key == "value":
            self.value = None
        elif key == "style":
            self.elem.styleDict = {}
            self.elem.send_state('styleDict')
        elif key == "javascript_handles":
            self.javascript_handles = None
        elif key == "event_handlers":
            self.event_handlers = None
        elif key == "onevents":
            self.onevents = None
        else:
            key = HTMLManager.clean_key(key)
            if key == "class":
                self.elem.classList = []
                self.elem.send_state('classList')
            elif key in CSS.known_properties:
                del self.elem.styleDict[key]
                self.elem.send_state('styleDict')
            else:
                del self.elem.elementAttributes[key]
                self.elem.send_state('elementAttributes')
    def get_child(self, position, wrapper=False):
        kids = self.elem.children
        if len(kids) == 0:
            body = self.html
            if body is None:
                raise IndexError("element {} has no body to index".format(self))
            val = body[position]
        else:
            val = kids[position]
        if wrapper and hasattr(val, 'wrappers'):
            return next(iter(val.wrappers))
        else:
            return val
    def set_child(self, position, value):
        kids = self.elem.children
        if len(kids) == 0:
            body = self.html
            if body is None:
                raise IndexError("element {} has no body to index".format(self))
            body[position] = value
        else:
            kids[position] = self.canonicalize_widget(value)
            self.children = kids
            # self.elem.children = kids
            # self.elem.send_state('children')
    def insert(self, where, child):
        kids = self.elem.children
        if len(kids) == 0:
            body = self.html
            if body is None:
                kids = list(kids)
                kids.insert(0, self.canonicalize_widget(child))
                self.children = tuple(kids)
            elif isinstance(child, (str, HTML.XMLElement)):
                body.insert(where, child)
            else:
                self.activate_body()
                self.insert(where, child)
        else:
            kids = list(kids)
            if where is None:
                where = len(kids)
            kids.insert(where, self.canonicalize_widget(child))
            self.children = tuple(kids)
    def append(self, child):
        self.insert(None, child)
    def del_child(self, position):
        kids = self.elem.children
        if len(kids) == 0:
            body = self.html
            if body is None:
                raise IndexError("element {} has no body to index".format(self))
            del body[position]
        else:
            del kids[position]
            self.children = kids
            # self.elem.children = kids
            # self.elem.send_state('children')
    def activate_body(self):
        html = self.html
        self.html_string = ""
        self.children = [self.canonicalize_widget(html)]
    @property
    def elements(self):
        kids = self.elem.children
        if len(kids) == 0:
            kids = []
            if self.text is not None:
                kids.append(self.text)
            elif self.html_string is not None:
                kids.append(self.html_string)
        return kids
    @property
    def children(self):
        return self.elem.children
    @children.setter
    def children(self, kids):
        self.elem.children = kids
        self.elem.send_state('children')
    @property
    def html_string(self):
        eid = self.elem.innerHTML
        return None if eid == "" else eid
    @html_string.setter
    def html_string(self, val):
        if len(self.elem.children) > 0:
            self.elem.children = []
        self.elem.innerHTML = val
        self._html_cache = None
    @property
    def html(self):
        if self._html_cache is None:
            self._html_cache = self.load_HTML()
        return self._html_cache
    @html.setter
    def html(self, html):
        self._html_cache = html
        self.html_string = html.tostring()
        self._html_cache = html
    def _track_html_change(self, *_):
        html = self._html_cache
        self.html_string = html.tostring()
        self._html_cache = html
    def load_HTML(self):
        html_base = self.html_string
        if html_base is not None:
            html_base = HTML.parse(html_base, strict=False)
            html_base.on_update = self._track_html_change
        return html_base

    @property
    def javascript_handles(self):
        if self._handle_api is None:
            return self.elem.jsHandlers
        else:
            return self._handle_api
    @javascript_handles.setter
    def javascript_handles(self, js):
        if hasattr(js, 'javascript_handles'):
            self._handle_api = js
            self.elem.jsAPI = js.elem
            self.elem.send_state('jsAPI')
        else:
            if js is not None and 'api' in js:
                js = js.copy()
                self._handle_api = js['api']
                self.elem.jsAPI = js['api'].elem
                self.elem.send_state('jsAPI')
                del js['api']
            else:
                self._handle_api = None
            self.elem.jsHandlers = js
            self.elem.send_state('jsHandlers')
    @property
    def class_list(self):
        return self.elem.classList
    @class_list.setter
    def class_list(self, cls):
        self.elem.classList = HTMLManager.manage_class(cls)
        self.elem.send_state('classList')
    def add_class(self, *cls):
        cl = self.elem.classList
        cur_len = len(cl)
        proper_classes = []
        for c in cls:
            proper_classes.extend(HTMLManager.manage_class(c))
        for c in proper_classes:
            if c not in cl:
                cl.append(c)
        if len(cl) > cur_len:
            self.elem.classList = cl
            self.elem.send_state('classList')
        return self
    def remove_class(self, *cls):
        cl = self.elem.classList
        cur_len = len(cl)
        proper_classes = []
        for c in cls:
            proper_classes.extend(HTMLManager.manage_class(c))
        for c in proper_classes:
            if c in cl:
                cl.remove(c)
        if len(cl) < cur_len:
            self.elem.classList = cl
            self.elem.send_state('classList')
        return self

    @property
    def style(self):
        return self.elem.styleDict
    @style.setter
    def style(self, style):
        self.elem.classList = HTMLManager.manage_styles(style).props
        self.elem.send_state('styleDict')
    def add_styles(self, **sty):
        sd = self.elem.styleDict
        for c,v in sty.items():
            sd[c] = v
        self.elem.styleDict = sd
        self.elem.send_state('styleDict')
    def remove_styles(self, *sty):
        sd = self.elem.styleDict
        for c in sty:
            if c in sd:
                del sd[c]
        self.elem.styleDict = sd
        self.elem.send_state('styleDict')

    @property
    def data(self):
        return self.elem.exportData
    @data.setter
    def data(self, d:dict):
        self.elem.exportData = d

    @property
    def event_handlers(self):
        return {
            k: self._event_handlers[k] if self._event_handlers[k] is not None else v
            for k, v in self.elem.eventPropertiesDict.items()
        }
    @event_handlers.setter
    def event_handlers(self, event_handlers):
        self.update_events(event_handlers)
    def update_events(self, events):
        self.add_event(**events, send=False)
        self.remove_event(*(k for k in self.elem.eventPropertiesDict if k not in events))
    def add_event(self, send=True, **events):
        ed = self.elem.eventPropertiesDict
        for c,v in events.items():
            if isinstance(v, dict):
                v = v.copy()
                if 'callback' in v:
                    callback = v['callback']
                    del v['callback']
                    self._event_handlers[c] = callback
                    v['notify'] = True
                else:
                    self._event_handlers[c] = None
                if len(v) > 0:
                    ed[c] = v
                else:
                    ed[c] = None
            elif isinstance(v, str):
                ed[c] = v
                self._event_handlers[c] = None
            else:
                ed[c] = None
                self._event_handlers[c] = v
        self.elem.eventPropertiesDict = ed
        if send:
            self.elem.send_state('eventPropertiesDict')
        if len(self.elem.eventPropertiesDict) > 0:
            self.elem.bind_callback(self.handle_event)
        return self
    def remove_event(self, *events, send=True):
        ed = self.elem.eventPropertiesDict
        for c in events:
            if c in ed:
                del ed[c]
            if c in self._event_handlers:
                del self._event_handlers[c]
        self.elem.eventPropertiesDict = ed
        if send:
            self.elem.send_state('eventPropertiesDict')
        if len(self.elem.eventPropertiesDict) == 0:
            self.elem.reset_callbacks()

    _message_waiting_semaphores = {}
    @staticmethod
    def _on_msg(msg, callback, current_event):
        def on_msg(e, self, msg=msg, callback=callback, current_event=current_event):
            res = None
            if callback is not None:
                res = callback(e, self)
            if current_event is not None:
                current_event()
            self._message_waiting_semaphores[msg] = [False, res]
        return on_msg
    async def _wait_for_message(self, msg, poll_interval=.05):
        # self._message_waiting_semaphores[msg] = [False, list(self.elem._callbacks.callbacks)]
        while self._message_waiting_semaphores.get(msg, [True, None])[0]:
            await asyncio.sleep(poll_interval)
    async def wait_for_message(self, msg, callback, suppress_others=False, timeout=1, poll_interval=.05):
        self._message_waiting_semaphores[msg] = [True, None]
        og_listener = self.event_handlers.get(msg, None)
        current_event = og_listener if not suppress_others else callback
        if not suppress_others:
            if isinstance(current_event, str):
                current_event = {'method':current_event}
            elif not isinstance(current_event, dict):
                current_event = {'callback':current_event}
            else:
                current_event = current_event.copy()
            current_event['callback'] = self._on_msg(msg, callback, current_event['callback'])
        self.add_event(
            **{msg:current_event}
        )
        # print(self._event_handlers)
        try:
            await asyncio.wait_for(asyncio.create_task(self._wait_for_message(msg, poll_interval=poll_interval)), timeout=timeout)
            res = self._message_waiting_semaphores[msg][1]
        finally:
            self._message_waiting_semaphores[msg][1] = None
            if og_listener is None:
                self.remove_event(msg)
            else:
                self.add_event(**{msg: og_listener})
        # del self._message_waiting_semaphores[msg]
        return res
    def call(self, method, buffers=None, return_message=None, callback=None, timeout=1, poll_interval=0.05, suppress_others=False, **content):
        if return_message is not None:
            fut = asyncio.ensure_future(
                self.wait_for_message(return_message, callback, timeout=timeout, poll_interval=poll_interval, suppress_others=suppress_others)
            )
        base = self.elem.call(method, content=content, buffers=buffers)
        if return_message is not None:
            return fut
        return base
    def _setup_thread_listener(self, msg, callback, suppress_others=False):
        self._message_waiting_semaphores[msg] = [True, None]
        og_listener = self.event_handlers.get(msg, None)
        current_event = og_listener if not suppress_others else callback
        if not suppress_others:
            if isinstance(current_event, str):
                current_event = {'method': current_event}
            elif not isinstance(current_event, dict):
                current_event = {'callback': current_event}
            else:
                current_event = current_event.copy()
            current_event['callback'] = self._on_msg(msg, callback, current_event['callback'])
        self.add_event(
            **{msg: current_event}
        )
    def _wait_for_thread_message(self, msg, poll_interval=.05):
        while self._message_waiting_semaphores.get(msg, [True, None])[0]:
            time.sleep(poll_interval)
    def _wait_for_result(self, msg, og_listener, timeout=1, poll_interval=.05):
        try:
            # just here to allow the coroutines to work while the thread thinks...
            t = threading.Thread(target=self._wait_for_thread_message, args=(msg,), kwargs=dict(poll_interval=poll_interval))
            t.start()
            t.join(timeout)
            if t.is_alive():
                self._message_waiting_semaphores[msg][0] = False
                raise TimeoutError("didn't get a result")
            res = self._message_waiting_semaphores[msg][1]
        finally:
            self._message_waiting_semaphores[msg][1] = None
            if og_listener is None:
                self.remove_event(msg)
            else:
                self.add_event(**{msg: og_listener})
        # del self._message_waiting_semaphores[msg]
        return res
    def _thread_call(self, method, buffers=None, return_message=None, callback=None, timeout=1, poll_interval=0.05, suppress_others=False, **content):
        if return_message is not None:
            og_listener = self._setup_thread_listener(return_message, callback, suppress_others=suppress_others)
        base = self.elem.call(method, content=content, buffers=buffers)
        if return_message is not None:
            return self._wait_for_result(return_message, og_listener, timeout=timeout, poll_interval=poll_interval)
        return base
    def add_javascript(self, **methods):
        ed = self.javascript_handles
        if isinstance(ed, ActiveHTMLWrapper):
            ed.add_javascript(**methods)
        else:
            ed.update(methods)
            self.javascript_handles = ed
        return self
    def remove_javascript(self, *methods):
        ed = self.javascript_handles
        if isinstance(ed, ActiveHTMLWrapper):
            ed.remove_javascript(*methods)
        else:
            for c in methods:
                if c in ed:
                    del ed[c]
            self.javascript_handles = ed
        return self

    def trigger(self, method, buffers=None, **content):
        return self.elem.trigger(method, content=content, buffers=buffers)
    @property
    def onevents(self):
        return {
            k:self._event_handlers[k] if self._event_handlers[k] is not None else v
            for k,v in self.elem.onevents.items()
        }
    @onevents.setter
    def onevents(self, onevents):
        self.update_onevents(onevents)
    def update_onevents(self, events):
        self.on(**events, send=False)
        self.off(*(k for k in self.elem.onevents if k not in events))
    def on(self, send=True, **events):
        ed = self.elem.onevents
        for c,v in events.items():
            if isinstance(v, dict):
                v = v.copy()
                if 'callback' in v:
                    callback = v['callback']
                    del v['callback']
                    self._event_handlers[c] = callback
                    v['notify'] = True
                else:
                    self._event_handlers[c] = None
                if len(v) > 0:
                    ed[c] = v
                else:
                    ed[c] = None
            elif isinstance(v, str):
                ed[c] = v
                self._event_handlers[c] = None
            else:
                ed[c] = None
                self._event_handlers[c] = v
        self.elem.onevents = ed
        if send:
            self.elem.send_state('onevents')
        if len(self.elem.onevents) > 0:
            self.elem.bind_callback(self.handle_event)
        return self
    def off(self, *events, send=True):
        ed = self.elem.onevents
        for c in events:
            if c in ed:
                del ed[c]
            if c in self._event_handlers:
                del self._event_handlers[c]
        self.elem.onevents = ed
        if send:
            self.elem.send_state('onevents')
        if len(self.elem.eventPropertiesDict) == 0:
            self.elem.reset_callbacks()

    @property
    def track_value(self):
        return self.elem.trackInput
    @track_value.setter
    def track_value(self, v):
        self.elem.trackInput = v
    @property
    def continuous_update(self):
        return self.elem.continuousUpdate
    @continuous_update.setter
    def continuous_update(self, v):
        self.elem.continuousUpdate = v

    class LazyLoader:
        def __init__(self, base_cls, args, kwargs):
            self.base = base_cls
            self.args = args
            self.kwargs = kwargs
            self._obj = None
        def load(self):
            if self._obj is None:
                self._obj = self.base(*self.args, **self.kwargs)
            return self._obj
    @classmethod
    def loader(cls, *args, **kwargs):
        return cls.LazyLoader(cls, args, kwargs)

class HTMLWidgets:

    @classmethod
    def get_exec_prefix(cls):
        import subprocess
        ext_call = subprocess.run(["jupyter", "labextension", "list"], capture_output=True)
        path_list = ext_call.stdout.decode() + "\n" + ext_call.stderr.decode()
        end_path = ("share", "jupyter", "labextensions")
        for p in path_list.split():
            if os.path.isdir(p):
                p, p3 = os.path.split(p)
                p, p2 = os.path.split(p)
                root, p1 = os.path.split(p)
                if (p1, p2, p3) == end_path: # TODO: be a bit more careful?
                    return root
        else:
            return sys.exec_prefix


    @classmethod
    def load(cls, exec_prefix=None, overwrite=False):
        from .ActiveHTMLWidget import HTMLElement
        if exec_prefix is None:
            exec_prefix = cls.get_exec_prefix()

        nb = HTMLElement.jupyternb_install(exec_prefix=exec_prefix, overwrite=overwrite)
        lab = HTMLElement.jupyterlab_install(exec_prefix=exec_prefix, overwrite=overwrite)
        res = []
        if nb is not None:
            res.append(nb)
        if lab is not None:
            res.append(lab)
        if len(res) > 0:
            from IPython.core.display import display
            display(*res)

    _cls_map = None
    @classmethod
    def get_class_map(cls):
        if cls._cls_map is None:
            cls._cls_map = {}
            for v in cls.__dict__.values():
                if isinstance(v, type) and hasattr(v, 'base') and hasattr(v.base, 'tag'):
                    cls._cls_map[v.base.tag] = v
        return cls._cls_map
    @classmethod
    def from_HTML(cls, html:HTML.XMLElement, event_handlers=None, debug_pane=None, **props):
        tag = html.tag
        map = cls.get_class_map()
        try:
            tag_class = map[tag]
        except KeyError:
            tag_class = ActiveHTMLWrapper#lambda *es,**ats:HTML.XMLElement(tag, *es, **ats)
        return tag_class.from_HTML(html, event_handlers=event_handlers, debug_pane=debug_pane, **props)

    class JavascriptAPI(ActiveHTMLWrapper):
        def __init__(self, safety_wrap=True, _debugPrint=False, disable_caching=True, **javascript_handles):
            if safety_wrap:
                javascript_handles = {
                    k:self.safety_wrap(v) if k not in {"api", "src"} else v
                    for k,v in javascript_handles.items()
                }
                if 'src' in javascript_handles and os.path.isfile(javascript_handles['src']): # need to relink to Jupyter path
                    absfile = os.path.abspath(javascript_handles['src'])
                    javascript_handles['src'] = "/files"+"/".join(pathlib.Path(absfile).parts)
            if disable_caching and 'src' in javascript_handles:
                import urllib.parse as parse
                scheme, netloc, path, params, query, fragment = parse.urlparse(javascript_handles['src'])
                query+="cachetime="+str(time.time())
                javascript_handles['src'] = parse.urlunparse((scheme, netloc, path, params, query, fragment))
            self._api_src=None
            super().__init__(javascript_handles=javascript_handles, _debugPrint=_debugPrint)
        safety_template="try{{\n{body}\n}} catch(error) {{ console.log(error); alert('An error occurred: ' + error.toString() + '; check console for details') }}"
        def safety_wrap(self, v):
            return self.safety_template.format(body=v)
    class WrappedHTMLElement(ActiveHTMLWrapper):
        def __repr__(self):
            body = (
                self.children if len(self.children) > 0
                else self.html_string if self.html_string is not None
                else self.text if self.text is not None
                else ""
            )
            return "{}Element({!r}, cls={}, style={}, id={})".format(
                type(self).__name__,
                body,
                self.class_list,
                self.style,
                self.id
            )
    class Abbr(WrappedHTMLElement): base=HTML.Abbr
    class Address(WrappedHTMLElement): base=HTML.Address
    class Anchor(WrappedHTMLElement): base=HTML.Anchor
    A = Anchor
    class Area(WrappedHTMLElement): base=HTML.Area
    class Article(WrappedHTMLElement): base=HTML.Article
    class Aside(WrappedHTMLElement): base=HTML.Aside
    class Audio(WrappedHTMLElement): base=HTML.Audio
    class B(WrappedHTMLElement): base=HTML.B
    class Base(WrappedHTMLElement): base=HTML.Base
    class BaseList(WrappedHTMLElement): base=HTML.BaseList
    class Bdi(WrappedHTMLElement): base=HTML.Bdi
    class Bdo(WrappedHTMLElement): base=HTML.Bdo
    class Blockquote(WrappedHTMLElement): base=HTML.Blockquote
    class Body(WrappedHTMLElement): base=HTML.Body
    class Bold(WrappedHTMLElement): base=HTML.Bold
    class Br(WrappedHTMLElement): base=HTML.Br
    class Button(WrappedHTMLElement): base=HTML.Button
    class Canvas(WrappedHTMLElement): base=HTML.Canvas
    class Caption(WrappedHTMLElement): base=HTML.Caption
    class Cite(WrappedHTMLElement): base=HTML.Cite
    # class ClassAdder(WrappedHTMLElement): base=HTMLManager.ClassAdder
    # class ClassRemover(WrappedHTMLElement): base=HTMLManager.ClassRemover
    class Code(WrappedHTMLElement): base=HTML.Code
    class Col(WrappedHTMLElement): base=HTML.Col
    class Colgroup(WrappedHTMLElement): base=HTML.Colgroup
    class Data(WrappedHTMLElement): base=HTML.Data
    class Datalist(WrappedHTMLElement): base=HTML.Datalist
    class Dd(WrappedHTMLElement): base=HTML.Dd
    class Del(WrappedHTMLElement): base=HTML.Del
    class Details(WrappedHTMLElement): base=HTML.Details
    class Dfn(WrappedHTMLElement): base=HTML.Dfn
    class Dialog(WrappedHTMLElement): base=HTML.Dialog
    class Div(WrappedHTMLElement): base=HTML.Div
    class Dl(WrappedHTMLElement): base=HTML.Dl
    class Dt(WrappedHTMLElement): base=HTML.Dt
    # class ElementModifier(WrappedHTMLElement): base=HTML.ElementModifier
    class Em(WrappedHTMLElement): base=HTML.Em
    class Embed(WrappedHTMLElement): base=HTML.Embed
    class Fieldset(WrappedHTMLElement): base=HTML.Fieldset
    class Figcaption(WrappedHTMLElement): base=HTML.Figcaption
    class Figure(WrappedHTMLElement): base=HTML.Figure
    class Footer(WrappedHTMLElement): base=HTML.Footer
    class Form(WrappedHTMLElement): base=HTML.Form
    class Head(WrappedHTMLElement): base=HTML.Head
    class Header(WrappedHTMLElement): base=HTML.Header
    class Heading(WrappedHTMLElement): base=HTML.Heading
    class Hr(WrappedHTMLElement): base=HTML.Hr
    class Html(WrappedHTMLElement): base=HTML.Html
    class Iframe(WrappedHTMLElement): base=HTML.Iframe
    class Image(WrappedHTMLElement): base=HTML.Image
    class Img(WrappedHTMLElement): base=HTML.Img
    class Input(WrappedHTMLElement): base=HTML.Input
    class Ins(WrappedHTMLElement): base=HTML.Ins
    class Italic(WrappedHTMLElement): base=HTML.Italic
    class Kbd(WrappedHTMLElement): base=HTML.Kbd
    class Label(WrappedHTMLElement): base=HTML.Label
    class Legend(WrappedHTMLElement): base=HTML.Legend
    class Li(WrappedHTMLElement): base=HTML.Li
    class Link(WrappedHTMLElement): base=HTML.Link
    class List(WrappedHTMLElement): base=HTML.List
    class ListItem(WrappedHTMLElement): base=HTML.ListItem
    class Main(WrappedHTMLElement): base=HTML.Main
    class Map(WrappedHTMLElement): base=HTML.Map
    class Mark(WrappedHTMLElement): base=HTML.Mark
    class Meta(WrappedHTMLElement): base=HTML.Meta
    class Meter(WrappedHTMLElement): base=HTML.Meter
    class Nav(WrappedHTMLElement): base=HTML.Nav
    class Noscript(WrappedHTMLElement): base=HTML.Noscript
    class NumberedList(WrappedHTMLElement): base=HTML.NumberedList
    class Object(WrappedHTMLElement): base=HTML.Object
    class Ol(WrappedHTMLElement): base=HTML.Ol
    class Optgroup(WrappedHTMLElement): base=HTML.Optgroup
    class Option(WrappedHTMLElement): base=HTML.Option
    class Output(WrappedHTMLElement): base=HTML.Output
    class P(WrappedHTMLElement): base=HTML.P
    class Param(WrappedHTMLElement): base=HTML.Param
    class Picture(WrappedHTMLElement): base=HTML.Picture
    class Pre(WrappedHTMLElement): base=HTML.Pre
    class Progress(WrappedHTMLElement): base=HTML.Progress
    class Q(WrappedHTMLElement): base=HTML.Q
    class Rp(WrappedHTMLElement): base=HTML.Rp
    class Rt(WrappedHTMLElement): base=HTML.Rt
    class Ruby(WrappedHTMLElement): base=HTML.Ruby
    class S(WrappedHTMLElement): base=HTML.S
    class Samp(WrappedHTMLElement): base=HTML.Samp
    class Script(WrappedHTMLElement): base=HTML.Script
    class Section(WrappedHTMLElement): base=HTML.Section
    class Select(WrappedHTMLElement): base=HTML.Select
    class Small(WrappedHTMLElement): base=HTML.Small
    class Source(WrappedHTMLElement): base=HTML.Source
    class Span(WrappedHTMLElement): base=HTML.Span
    class Strong(WrappedHTMLElement): base=HTML.Strong
    class Style(WrappedHTMLElement): base=HTML.Style
    # class StyleAdder(WrappedHTMLElement): base=HTML.StyleAdder
    class Sub(WrappedHTMLElement): base=HTML.Sub
    class SubHeading(WrappedHTMLElement): base=HTML.SubHeading
    class SubsubHeading(WrappedHTMLElement): base=HTML.SubsubHeading
    class SubsubsubHeading(WrappedHTMLElement): base=HTML.SubsubsubHeading
    class SubHeading5(WrappedHTMLElement): base=HTML.SubHeading5
    class SubHeading6(WrappedHTMLElement): base=HTML.SubHeading6
    class Summary(WrappedHTMLElement): base=HTML.Summary
    class Sup(WrappedHTMLElement): base=HTML.Sup
    class Svg(WrappedHTMLElement): base=HTML.Svg
    class Table(WrappedHTMLElement): base=HTML.Table
    class TableHeader(WrappedHTMLElement): base=HTML.TableHeader
    class TableBody(WrappedHTMLElement): base=HTML.TableBody
    class TableHeading(WrappedHTMLElement): base=HTML.TableHeading
    class TableItem(WrappedHTMLElement): base=HTML.TableItem
    class TableRow(WrappedHTMLElement): base=HTML.TableRow
    class TagElement(WrappedHTMLElement): base=HTML.TagElement
    class Tbody(WrappedHTMLElement): base=HTML.Tbody
    class Td(WrappedHTMLElement): base=HTML.Td
    class Template(WrappedHTMLElement): base=HTML.Template
    class Text(WrappedHTMLElement): base=HTML.Text
    class Textarea(WrappedHTMLElement): base=HTML.Textarea
    class Tfoot(WrappedHTMLElement): base=HTML.Tfoot
    class Th(WrappedHTMLElement): base=HTML.Th
    class Thead(WrappedHTMLElement): base=HTML.Thead
    class Time(WrappedHTMLElement): base=HTML.Time
    class Title(WrappedHTMLElement): base=HTML.Title
    class Tr(WrappedHTMLElement): base=HTML.Tr
    class Track(WrappedHTMLElement): base=HTML.Track
    class U(WrappedHTMLElement): base=HTML.U
    class Ul(WrappedHTMLElement): base=HTML.Ul
    class Var(WrappedHTMLElement): base=HTML.Var
    class Video(WrappedHTMLElement): base=HTML.Video
    class Wbr(WrappedHTMLElement): base=HTML.Wbr

    class OutputArea(ActiveHTMLWrapper):
        def __init__(self, *elements, max_messages=None, autoclear=False, event_handlers=None, cls=None, **styles):
            if len(elements) == 1 and isinstance(elements, (list, tuple)):
                elements = elements[0]
            self._call_depth = 0
            self.autoclear = autoclear
            self.max_messages = max_messages
            self.output = JupyterAPIs.get_widgets_api().Output()
            elements = list(elements) + [self.output]
            super().__init__(*elements, event_handlers=event_handlers, cls=cls, **styles)
        def print(self, *args, **kwargs):
            with self:
                print(*args, **kwargs)
        def show_output(self, *args, **kwargs):
            with self:
                return JupyterAPIs().display_api.display(*args, **kwargs)

        def _get_display_data(self, args):
            outs = []
            for output in args:
                # Jupyter Widgets is broken so we need to patch...
                if hasattr(output, 'savefig'):
                    buf = io.BytesIO()
                    output.savefig(buf, format='png')
                    output = JupyterAPIs.get_display_api().Image(buf.getvalue())
                if output is not None:
                    if hasattr(output, 'get_mime_bundle'):
                        outs.append({
                            'output_type': 'display_data',
                            'data': output.get_mime_bundle(),
                            'metadata': {}
                        })
                    else:
                        if isinstance(output, str):
                            outs.append({
                                'output_type': 'display_data',
                                'data': {'text/plain':output},
                                'metadata': {}
                            })
                        else:
                            fmt = JupyterAPIs.get_shell_instance().display_formatter.format
                            data, metadata = fmt(output)
                            outs.append({
                                'output_type': 'display_data',
                                'data': data,
                                'metadata': metadata
                            })
            return outs

        def append_stdout(self, arg):
            self.output.append_stdout(arg)
        def append_stderr(self, arg):
            self.output.append_stderr(arg)
        def show_buffered(self, *args):
            self.output.outputs += tuple(self._get_display_data(args))
        def set_output(self, *args):
            # with self.output:
            self.output.outputs = tuple(self._get_display_data(args))
        def show_raw(self, *args):
            with self:
                api = JupyterAPIs().display_api
                api.publish_display_data(*({'text/plain': t} if isinstance(t, str) else t for t in args))
                # api.display(*(api.TextDisplayObject(t) for t in args))
        def clear(self, wait=False):
            self.output.clear_output(wait=wait)
            self.output.outputs = ()
        def __enter__(self):
            if self._call_depth < 1 and self.autoclear:
                self.clear(wait=True)
            self._call_depth += 1
            self.output.__enter__()
            return self
        def __exit__(self, exc_type, exc_val, exc_tb):
            self._call_depth = max(self._call_depth-1, 0)
            if self.max_messages is not None:
                n = self.max_messages
                if len(self.output.outputs) > n:
                    self.output.outputs = self.output.outputs[-n:]
                elif len(self.output.outputs) == 1:  # and isinstance(self.output.outputs[0], str):
                    out_dict = self.output.outputs[0].copy()
                    msg = out_dict['text']
                    if msg.count('\n') > n:
                        msg = "\n".join(msg.splitlines()[-n:]) + "\n"
                        out_dict['text'] = msg
                    with self.output:
                        self.output.outputs = (out_dict,)
                        self.output._flush()
            return self.output.__exit__(exc_type, exc_val, exc_tb)



import enum, numpy as np, collections

__all__ = [
    "FFIModule",
    "FFIMethod",
    "FFIArgument",
    "FFIType"
]

try: # guard clause
    np_float128 = np.float128
except AttributeError:
    np_float128 = np.longdouble # bad vibes but we'll see what happens

class FFIType(enum.Enum):
    """
    The set of supported enum types.
    Maps onto the native python convertable types and NumPy dtypes.
    In the future, this should be done more elegantly, but for now it suffices
    that these types align on the C++ side and this side.
    Only NumPy arrays are handled using the buffer interface & so if you want to pass a pointer
    you gotta do it using a NumPy array.
    """

    _type_map = {}

    GENERIC = -1

    Void = 1
    _type_map[Void] = ("void", None)

    # for k in type_map[NUMPY_Float64]: type_map[k] = NUMPY_Float64

    PY_TYPES=1000

    UnsignedChar = PY_TYPES + 10
    _type_map[UnsignedChar] = ("b", int)
    Short = PY_TYPES + 20
    _type_map[Short] = ("h", int)
    UnsignedShort = PY_TYPES + 21
    _type_map[UnsignedShort] = ("H", int)
    Int = PY_TYPES + 30
    _type_map[Int] = ("i", int)
    UnsignedInt = PY_TYPES + 31
    _type_map["I"] = UnsignedInt
    _type_map[UnsignedInt] = ("I", int)
    Long = PY_TYPES + 40
    _type_map[Long] = ("l", int)
    UnsignedLong = PY_TYPES + 41
    _type_map[UnsignedLong] = ("L", int)
    LongLong = PY_TYPES + 50
    _type_map["k"] = LongLong
    _type_map[LongLong] = ("k", int)
    UnsignedLongLong = PY_TYPES + 51
    _type_map["K"] = UnsignedLongLong
    _type_map[UnsignedLongLong] = ("K", int)
    PySizeT = PY_TYPES + 60
    _type_map[PySizeT] = ("n", int)

    Float = PY_TYPES + 70
    _type_map["f"] = Float
    _type_map[Float] = ("f", float)
    Double = PY_TYPES + 71
    _type_map["d"] = Double
    _type_map[Double] = ("d", float)

    Bool = PY_TYPES + 80
    _type_map[Bool] = ("p", bool)
    String = PY_TYPES + 90
    _type_map[String] = ("s", str)
    PyObject = PY_TYPES + 100
    _type_map["O"] = PyObject
    _type_map[PyObject] = ("O", object)

    Compound = PY_TYPES + 500
    _type_map[Compound] = ("compound", dict)

    # supports the NumPy NPY_TYPES enum
    # 200 is python space
    NUMPY_TYPES = 2000

    NUMPY_Int8 = NUMPY_TYPES + 10
    _type_map[NUMPY_Int8] = ("int8", np.int8)
    NUMPY_UnsignedInt8 = NUMPY_TYPES + 11
    _type_map[NUMPY_UnsignedInt8] = ("uint8", np.uint8)
    NUMPY_Int16 = NUMPY_TYPES + 12
    _type_map[NUMPY_Int16] = ("int16", np.int16)
    NUMPY_UnsignedInt16 = NUMPY_TYPES + 13
    _type_map[NUMPY_UnsignedInt16] = ("uint16", np.uint16)
    NUMPY_Int32 = NUMPY_TYPES + 14
    _type_map[NUMPY_Int32] = ("int32", np.int32)
    NUMPY_UnsignedInt32 = NUMPY_TYPES + 15
    _type_map[NUMPY_UnsignedInt32] = ("uint32", np.uint32)
    NUMPY_Int64 = NUMPY_TYPES + 16
    _type_map[NUMPY_Int64] = ("int64", np.int64)
    NUMPY_UnsignedInt64 = NUMPY_TYPES + 17
    _type_map[NUMPY_UnsignedInt64] = ("uint64", np.uint64)

    NUMPY_Float16 = NUMPY_TYPES + 20
    _type_map[NUMPY_Float16] = ("float16", np.float16)
    NUMPY_Float32 = NUMPY_TYPES + 21
    _type_map[NUMPY_Float32] = ("float32", np.float32)
    NUMPY_Float64 = NUMPY_TYPES + 22
    _type_map[NUMPY_Float64] = ("float64", np.float64)
    NUMPY_Float128 = NUMPY_TYPES + 23
    _type_map[NUMPY_Float128] = ("float128", np_float128)

    NUMPY_Bool = NUMPY_TYPES + 30
    _type_map[NUMPY_Bool] = ("bool", bool)

    @classmethod
    def type_data(cls, val):
        mapp = cls._type_map.value
        if isinstance(val, cls):
            val = val.value
        return mapp[val]

    _rev_map = {}
    @classmethod
    def resolve_ffi_type(cls, val):
        mapp = cls._rev_map.value
        if len(mapp) == 0: # initialize reverse map
            for k,v in cls._type_map.value.items():
                if isinstance(v, tuple):
                    for kk in v:
                        mapp[kk] = k
        return FFIType(mapp[val])

class FFIContainerType(enum.Enum):
    Untyped = 0
    Raw = 1
    Vector = 2
    Array = 3

class DebugLevels(enum.Enum):
    Quiet = 0
    Normal = 5
    More = 10
    All = 50
    Excessive = 100

class ThreadingMode(enum.Enum):
    Serial = 'serial'
    OpenMP = 'omp'
    TBB = 'tbb'


class FFISpec:
    """
    Provides a uniform layout for handling specs of different parts of an FFI library
    """
    __fields__ = []

    def __init__(self, **kwargs):
        for k in self.__fields__:
            if k not in kwargs or kwargs[k] is None:
                raise ValueError("{} got no value for required parameter {}".format(
                    type(self).__name__,
                    k
                ))
        for k in kwargs:
            if k not in self.__fields__:
                raise ValueError("{} got value for unexpected parameter {}".format(
                    type(self).__name__,
                    k
                ))
class FFIArgument(FFISpec):
    """
    An argument spec for data to be passed to an FFIMethod
    """
    __fields__ = ["name", "dtype", "shape", "container_type"]
    def __init__(self, name=None, dtype=None, shape=None, container_type=None, value=None):
        if shape is None:
            shape = ()
        if container_type is None:
            container_type = 'Untyped'
        super().__init__(name=name, dtype=dtype, shape=shape, container_type=container_type)
        self.arg_name = name
        self.arg_type = self.infer_dtype(dtype)
        self.arg_shape = shape
        self.container_type = self.infer_ctype(container_type)
    _base_dtype_map = {
        'int8':FFIType
    }
    @classmethod
    def infer_dtype(cls, dtype):
        if dtype is None:
            raise ValueError("can't infer dtype for value `None`")
        if isinstance(dtype, FFIType):
            return dtype
        elif isinstance(dtype, (int, np.integer)):
            return FFIType(dtype)
        elif isinstance(dtype, str):
            try:
                return FFIType[dtype]
            except KeyError:
                return FFIType.resolve_ffi_type(dtype)
        elif isinstance(dtype, np.dtype):
            return FFIType.resolve_ffi_type(dtype.name)
        else:
            return FFIType.resolve_ffi_type(dtype)
            # for type_val, pair in FFIType.type_map.value.items():
            #         if isinstance(pair, tuple) and pair[1] == dtype:
            #             return FFIType(type_val)
            # raise ValueError("can't infer FFIType for {}".format(dtype))
    @classmethod
    def infer_ctype(cls, container_type):
        if isinstance(container_type, FFIContainerType):
            return container_type
        elif isinstance(container_type, str):
            return FFIContainerType[container_type]
        else:
            return FFIContainerType(container_type)
    @classmethod
    def from_arg_sig(cls, arg):
        #arg is an ArgumentSignature.Argument but I need to do a more serious
        # refactor later so I won't import that module yet
        return cls(
            arg.name,
            arg.typechar,
            shape=None, # no shapes built into Argument at this point,
            container_type=FFIContainerType.Raw if (arg.is_pointer() or arg.is_array()) else None
        )
    def __repr__(self):
        return "{}('{}', {}, {}, {})".format(
            type(self).__name__,
            self.arg_name,
            self.arg_type,
            self.arg_shape,
            self.container_type,
        )
    def cast(self, val):
        """

        :param val:
        :type val:
        :return:
        :rtype:
        """

        type_str, dtype = FFIType.type_data(self.arg_type)
        if isinstance(dtype, np.dtype): # have a numpy type
            dat = np.asanyarray(val, dtype=dtype)
        elif len(self.arg_shape) > 0:
            dat = np.asanyarray(val, dtype=dtype)
        else:
            if isinstance(val, np.ndarray):
                dat = np.ascontiguousarray(val.astype(dtype))
            elif not isinstance(val, dtype):
                dat = dtype(val)
            else:
                dat = val

        # do some shape checks...
        return FFIParameter(self, dat)
class FFIParameter:
    """
    Just an FFIArgument + associated value
    """
    def __init__(self, arg, val):
        if not isinstance(arg, FFIArgument):
            raise TypeError("{}: arg is expected to be an FFIArgument")
        self.arg = arg
        self.val = val
    @property
    def arg_name(self):
        return self.arg.arg_name
    @property
    def arg_type(self):
        return self.arg.arg_type
    @property
    def arg_shape(self):
        argshp = self.arg.arg_shape
        if isinstance(self.val, np.ndarray) and (0 in argshp or len(argshp) == 0):
            argshp = self.val.shape # gotta be a numpy value
        return argshp
    @property
    def container_type(self):
        return self.arg.container_type
    @property
    def arg_value(self):
        if (
                len(self.arg_shape) > 0
                or
                self.container_type == FFIContainerType.Raw
                or
                self.container_type == FFIContainerType.Vector
        ) and not isinstance(self.val, np.ndarray):
            self.val = np.ascontiguousarray(np.array(self.val)) # to be able to treat as a pointer
            # raise Exception(self.val)
            # if val.shape == ():
            #     val = np.array([self.val])
            # self.val = val
        return self.val
    def __repr__(self):
        return "{}({}, {}, {})->{}".format(
            type(self).__name__,
            self.arg_name,
            self.arg_type,
            self.arg_shape,
            self.arg_value
        )

class FFIParameters:
    """

    """
    def __init__(self, dats):
        self.dats = dats
        self._params = None
        self._map = None
    def __iter__(self):
        return iter(self.ffi_parameters)
    @property
    def ffi_parameters(self):
        if self._params is None:
            if isinstance(self.dats, (dict, collections.OrderedDict)):
                dat = tuple(self.dats.values())
            else:
                dat = self.dats
            self._params = dat # trying to keep stuff from going out of scope?
        return self._params
    @property
    def ffi_map(self):
        if self._map is None:
            self._map = {}
            for p in self.ffi_parameters:
                self._map[p.arg_name] = p
        return self._map
    def __getitem__(self, item):
        return self.ffi_map[item].arg_value

class FFIMethod(FFISpec):
    """
    Represents a C++ method callable through the plzffi interface
    """
    __fields__ = ["name", "arguments", "rtype", "vectorized"]
    def __init__(self, name=None, arguments=None, rtype=None, vectorized=None, module=None):
        super().__init__(name=name, arguments=arguments, rtype=rtype, vectorized=vectorized)
        self.name = name
        self.args = [FFIArgument(**x) if not isinstance(x, FFIArgument) else x for x in arguments]
        self.rtype = FFIType(rtype)
        self.vectorized = vectorized
        self.mod = module #type: None | FFIModule
    def bind_module(self, mod):
        self.mod = mod #type: FFIModule

    @property
    def arg_names(self):
        return tuple(x.arg_name for x in self.args)

    @classmethod
    def collect_args_from_list(cls, arg_list, *args, excluded_args=None, **kwargs):
        arg_dict = collections.OrderedDict()

        req_dict = collections.OrderedDict(
            (k.arg_name, k)
            for k in arg_list
        )

        for k in kwargs:
            arg_dict[k] = req_dict[k].cast(kwargs[k])
            del req_dict[k]

        if excluded_args is not None:
            for k in excluded_args:
                if k in req_dict:
                    del req_dict[k]

        if len(req_dict) > 0:
            for v, k in zip(args, req_dict.copy()):
                arg_dict[k] = req_dict[k].cast(v)
                del req_dict[k]

        if len(req_dict) > 0:
            raise ValueError("{}.{}: missing required arguments {}".format(
                cls.__name__,
                'collect_args',
                tuple(req_dict.values())
            ))
        return arg_dict

    def collect_args(self, *args, excluded_args=None, **kwargs):
        return self.collect_args_from_list(self.args, *args, excluded_args=excluded_args, **kwargs)

    @classmethod
    def from_signature(cls, sig, module=None):
        name, args, ret, vec = sig
        return cls(
            name=name,
            arguments=[{k:v for k,v in zip(FFIArgument.__fields__, x)} for x in args],
            rtype=ret,
            vectorized=vec,
            module=module
        )

    def call(self, *args, debug=False, **kwargs):
        fack = self.collect_args(*args, **kwargs)
        return self.mod.call_method(self.name, fack, debug=debug)
    def call_threaded(self, *args, threading_var=None, threading_mode="serial", debug=False, **kwargs):
        fack = self.collect_args(*args, **kwargs)
        return self.mod.call_method_threaded(self.name, fack, threading_var, mode=threading_mode, debug=debug)
    def __call__(self, *args, threading_var=None, threading_mode="serial", debug=False, **kwargs):
        fack = self.collect_args(*args, **kwargs)
        if threading_var is None and (threading_mode is None or threading_mode=='serial'):
            return self.mod.call_method(self.name, fack, debug=debug)
        else:
            return self.mod.call_method_threaded(self.name, fack, threading_var, mode=threading_mode, debug=debug)

    def __repr__(self):
        return "{}('{}', {})=>{}".format(
            type(self).__name__,
            self.name,
            self.args,
            "array({})".format(self.rtype) if self.vectorized else self.rtype
        )

class FFIModule(FFISpec):
    """
    Provides a layer to ingest a Python module containing an '_FFIModule' capsule.
    The capsule is expected to point to a `plzffi::FFIModule` object and can be called using a `PotentialCaller`
    """
    __fields__ = ["name", "methods"]
    def __init__(self, name=None, methods=None, module=None):
        super().__init__(name=name, methods=methods)
        self.name = name
        self.methods = [FFIMethod(**x) if not isinstance(x, FFIMethod) else x for x in methods]
        for m in self.methods:
            m.bind_module(self)
        self.mod = module

    @property
    def captup(self):
        return self.mod._FFIModule

    @classmethod
    def from_lib(cls, name,
                 src=None,
                 threaded=None,
                 extra_compile_args=None,
                 extra_link_args=None,
                 linked_libs=None,
                 **compile_kwargs
                 ):
        from .Loader import FFILoader
        if threaded is not None: # use the default
            compile_kwargs['threaded'] = threaded
        return FFILoader(name,
                         src=src,
                         extra_compile_args=extra_compile_args,
                         extra_link_args=extra_link_args,
                         linked_libs=linked_libs,
                         **compile_kwargs
                         ).call_obj

    @classmethod
    def from_signature(cls, sig, module=None):
        name, meths = sig
        return cls(
            name=name,
            methods=[FFIMethod.from_signature(x) for x in meths],
            module=module
        )
    @classmethod
    def get_debug_level(cls, debug):
        if debug is False:
            debug = DebugLevels.Quiet
        elif debug is True:
            debug = DebugLevels.Normal
        elif isinstance(debug, str):
            debug = DebugLevels[debug.lower().capitalize()]
        elif not isinstance(debug, (int, float, np.integer, np.floating)):
            debug = DebugLevels(debug)
        if isinstance(debug, DebugLevels):
            debug = debug.value
        return debug
    @classmethod
    def from_module(cls, module, debug=False):
        sig = module.get_signature(module._FFIModule, cls.get_debug_level(debug))
        return cls.from_signature(sig, module=module)

    @property
    def method_names(self):
        return tuple(x.name for x in self.methods)

    def get_method(self, name):
        # print(self.method_names)
        try:
            idx = self.method_names.index(name)
        except (ValueError, IndexError):
            idx = None
        if idx is not None:
            return self.methods[idx]
        else:
            raise AttributeError("FFIModule {} has no method {}".format(self.name, name))

    def call_method(self, name, params, debug=False):
        """
        Calls a method

        :param name:
        :type name:
        :param params:
        :type params:
        :return:
        :rtype:
        """
        req_attrs = ("arg_type", "arg_name", "arg_shape", "arg_value")
        if isinstance(params, (dict, collections.OrderedDict)):
            params = FFIParameters(params)
        for p in params:
            if not all(hasattr(p, x) for x in req_attrs):
                raise AttributeError("parameter {} needs attributes {}".format(p, req_attrs))
        return self.mod.call_method(self.captup, name, params, self.get_debug_level(debug))
    def call_method_threaded(self, name, params, thread_var, mode="serial", debug=False):
        """
        Calls a method with threading enabled

        :param name:
        :type name:
        :param params:
        :type params:
        :param thread_var:
        :type thread_var: str
        :param mode:
        :type mode:
        :return:
        :rtype:
        """
        req_attrs = ("arg_type", "arg_name", "arg_shape", "arg_value")
        if isinstance(params, (dict, collections.OrderedDict)):
            params = FFIParameters(params)
        for p in params:
            if not all(hasattr(p, x) for x in req_attrs):
                raise AttributeError("parameter needs attributes {}", req_attrs)
        if not isinstance(mode, ThreadingMode):
            mode = ThreadingMode(mode)
        mode = mode.name
        return self.mod.call_method_threaded(self.captup, name, params, thread_var, mode, self.get_debug_level(debug))

    def __getattr__(self, item):
        return self.get_method(item)

    def __repr__(self):
        return "{}('{}', methods={})".format(
            type(self).__name__,
            self.name,
            self.method_names
        )


"""
Defines a common data handler
"""
import os, sys

__all__ = [ "DataHandler", "DataError", "DataRecord" ]

default_data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
default_data_package = "TheRealMcCoy"
default_data_key = "data"
default_data_source_key = "source"

class DataError(KeyError):
    """
    Exception subclass for data error
    """

class DataHandler:
    """
    Defines a general data loader class that we can use for `AtomData` and any other data classes we might find useful.
    """
    def __init__(self,
                 data_name,
                 data_key=None,
                 source_key=None,
                 data_dir=None,
                 data_pkg=None,
                 alternate_keys=None,
                 getter=None,
                 record_type=None
                 ):
        """
        :param data_name: the name of the dataset
        :type data_name: str
        :param data_key: the key in the loaded dictionary to use for the actual data (`"data"` by default)
        :type data_key: str | None
        :param source_key: the key in the loaded dictionary for the original data source (`"source"` by default)
        :type source_key: str | None
        :param data_dir: the main directory data will be loaded from (`.` by default)
        :type data_dir: str | None
        :param data_pkg: the python package to load (`TheRealMcCoy` by default)
        :type data_pkg: str | None
        :param alternate_keys: alternate keys that can be used to index into the dataset which can will be populated at runtime
        :type alternate_keys: Iterable[str] | None
        :param getter: a function to use to resolve a key
        :type getter: callable | None
        :param record_type: the class to use for holding data (`DataRecord` by default)
        :type record_type: type | None
        """
        if data_dir is None:
            data_dir = default_data_dir
        if data_key is None:
            data_key = default_data_key
        if source_key is None:
            source_key = default_data_source_key
        if data_pkg is None:
            data_pkg = default_data_package
        self._data = None # this'll be a dict where we store all our data
        self._src = None
        self._dir = data_dir
        self._name = data_name
        self._key = data_key
        self._src_key = source_key
        self._pkg = data_pkg
        self._alts = alternate_keys
        self.record_type = DataRecord if record_type is None else record_type
        self._loaded = False
        self.getter = getter
    @property
    def data_file(self): # in case other people want to load it...
        return os.path.join(self._dir, self._name+".py")
    def _load_alts(self):
        # assumes we have dict data, but this entire structure does that anyway
        if not self._alts is None:
            extras = {}
            if isinstance(self._alts, str):
                self._alts = (self._alts,)
            for k in self._alts:
                extras.update({a[k]:a for a in self._data.values()}) #shouldn't increase memory bc mutable
            self._data.update(extras)
    def load(self, env=None):
        """
        Actually loads the data from `data_file`.
        Currently set up to just use an `import` statement but should
        be reimplemented to use a `Deserializer` from `Scaffolding.Serializers`

        :return:
        :rtype:
        """
        # currently we only load python data
        # TODO: I should rewrite this to use a Deserializer object...
        env = {} if env is None else env
        import sys
        sys.path.insert(0, self._dir) #name needs to be unique enough...
        if self._pkg is None:
            exec_stmt = "from {0} import {1} as {1}".format(self._name, self._key)
        else:
            exec_stmt = "from {0}.{1} import {2} as {2}".format(self._pkg, self._name, self._key)
        exec(exec_stmt, env, env)
        self._data = env[self._key]
        try:
            self._src = env[self._src_key]
        except:
            pass
        self._load_alts()
        self._loaded = True
    @property
    def data(self):
        if self._data is None:
            self.load()
        return self._data
    @property
    def source(self):
        if self._src is None and not self._loaded:
            self.load()
        return self._src
    def _get_data(self, key):
        def _get(a, k):
            if k not in a:
                raise DataError("{}: data source {} doesn't have subkey {} of {}".format(
                    type(self).__name__,
                    self._name,
                    k,
                    key
                ))
            return a[k]

        data = self.data
        if self.getter is None:
            if isinstance(key, tuple):
                from functools import reduce
                return reduce(_get, key, data)
            else:
                if key not in data:
                    raise DataError("{}: data source {} doesn't have key {}".format(
                        type(self).__name__,
                        self._name,
                        key
                    ))
                return data[key]
        else:
            return self.getter(data, key)
    def __getitem__(self, key):
        data = self._get_data(key)
        return self.record_type(self, key, data)
    def __len__(self):
        return len(self.data)
    def __iter__(self):
        return iter(self.data.items())

    # implementing to make pickling of data objects possible...
    def __getstate__(self):
        state = self.__dict__.copy()
        state['_data'] = None
        state['_src'] = None
        state['_loaded'] = False

        # raise Exception(state)
        return state
    def __setstate__(self, state):
        self.__dict__.update(state)
        self._loaded = False

    def __repr__(self):
        return "{}('{}', file='{}')".format(
            type(self).__name__,
            self._name,
            self._src
        )

class DataRecord:
    """
    Represents an individual record that might be accessed from a `DataHandler`.
    Implements _most_ of the `dict` interface, but, to make things a bit easier when
    pickling, is not implemented as a proper subclass of `dict`.
    """
    def __init__(self, data_handler, key, records):
        self.data = records
        self.handler = data_handler
        self.key = key

    def keys(self):
        return self.data.keys()
    def values(self):
        return self.data.keys()
    def items(self):
        return self.data.keys()

    def __getitem__(self, item):
        return self.data[item]

    def __repr__(self):
        return "{}('{}', {})".format(
            type(self).__name__,
            self.key,
            self.handler
        )

    # implementing to make pickling of data objects possible...
    def __getstate__(self):
        # it turns out we really need the dict.copy()...?
        state = self.__dict__.copy()
        del state['data']
        return state
    def __setstate__(self, state):
        self.__dict__.update(state)
        self.data = self.handler._get_data(self.key)




import abc, numpy as np, typing, weakref, hashlib
import collections

from ...Numputils import vec_outer, vec_tensordot
from ...Combinatorics import IntegerPartitioner, UniquePermutations

__all__ = [
    'TensorDerivativeConverter',
    'TensorExpansionTerms',
    'TensorExpression'
]


class TensorExpression:
    def __init__(self, expr:'TensorExpression.Term|List', **vars):
        self.expr = expr
        self.vars = vars
        self._prims = None
    def eval(self, subs:dict=None, print_terms=False):
        if subs is None:
            subs = self.vars
        if subs is None:
            subs = {}
        prims = self.primitives
        cache = {}
        names = {p.name:p for p in prims}
        _ac = TensorExpression.Term._array_cache
        TensorExpression.Term._array_cache = {k:v for k,v in zip(_ac.keys(), _ac.values())}
        for k in subs.keys() & names.keys():
            p = names[k]
            v = subs[k]
            cache[p] = TensorExpression.Term._array_cache.get(p, None)
            TensorExpression.Term._array_cache[p] = v

        try:
            if isinstance(self.expr, TensorExpression.Term):
                res = self.expr.asarray(print_terms=print_terms)
                if isinstance(res, TensorExpression.ArrayStack):
                    res = res.array
            else: # depth-first evaluation
                dfs_queue = collections.deque([self.expr])
                stack = collections.deque()
                sentinel = object() # for reconstructing arrays
                res = None # by construction we should never use this...?
                while dfs_queue:
                    e = dfs_queue.popleft()
                    # print("<<<<", id(e), str(e)[:25].replace("\n", ""))
                    if isinstance(e, TensorExpression.Term):
                        ar = e.asarray(print_terms=print_terms)
                        if isinstance(ar, TensorExpression.ArrayStack):
                            ar = ar.array
                        # print(ar.shape)
                        res.append(ar)
                    elif e is sentinel:
                        # print("  ="*10)
                        new_res = stack.pop()
                        if new_res is not None: # hit the bottom of the stack
                            new_res.append(res)
                            res = new_res
                        elif dfs_queue:
                            raise ValueError("sentinel at bad time")
                        else:
                            break
                    else:
                        stack.append(res)
                        res = []
                        # stick a sentinel in so we know when we've finished the array and can pop back out
                        dfs_queue.appendleft(sentinel)
                        # for sub_e in reversed(e):
                            # print("   >>>>", id(sub_e), str(sub_e)[:25].replace("\n", ""))
                        dfs_queue.extendleft(reversed(e))

        finally:
            TensorExpression.Term._array_cache = _ac
            # for p,v in cache.items():
            #     if v is None:
            #         del TensorExpression.Term._array_cache[p]
            #     else:
            #         TensorExpression.Term._array_cache[p] = v
        return res
    @property
    def primitives(self):
        if self._prims is None:
            self._prims = self.get_prims()
        return self._prims
    def walk(self, callback):
        bfs_queue = collections.deque([self.expr])
        visited = set()
        while bfs_queue:
            e = bfs_queue.popleft()
            if isinstance(e, TensorExpression.Term):
                if e not in visited:
                    bfs_queue.extend(c for c in e.children if isinstance(c, TensorExpression.Term))
                    callback(e)
                    visited.add(e)
            else:
                bfs_queue.extend(e) # flattening arrays
    def get_prims(self):
        kids = set()
        def add_if_prim(e):
            if isinstance(e, TensorExpression.Term) and len(e.children) == 0:
                kids.add(e)
        self.walk(add_if_prim)
        return kids
    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self.expr)

    class ArrayStack:
        __slots__ = ['stack_shape', 'array']
        def __init__(self, shape, array):
            """
            :param shape: shape of the stack of arrays to thread over
            :type shape:
            :param array:
            :type array:
            """
            self.array = array
            self.stack_shape = shape
        def __hash__(self):
            buf = self.array.view(np.uint8)
            return hash(hashlib.sha1(buf).hexdigest())
        def __eq__(self, other):
            return (
                    type(other) is type(self)
                    and self.stack_shape == other.stack_shape
                    and np.all(self.array == other.array)
            )
        @staticmethod
        def _make_broadcastable(a1, a2, offset):
            if a1.ndim < a2.ndim:
                for _ in range(a2.ndim - a1.ndim):
                    a1 = np.expand_dims(a1, -1)
            elif a2.ndim < a1.ndim:
                for _ in range(a1.ndim - a2.ndim):
                    a2 = np.expand_dims(a2, offset)
            return a1, a2
        @classmethod
        def _pad_shape(cls, arr, other, stack_shape):
            ogs = other.shape
            for _ in stack_shape:
                other = np.expand_dims(other, 0)
            other = np.broadcast_to(other, stack_shape + ogs)
            return cls._make_broadcastable(arr, other, len(stack_shape))
        def __add__(self, other):
            if isinstance(other, TensorExpression.ArrayStack):
                a1, a2 = self._make_broadcastable(
                    self.array,
                    other.array,
                    other.stack_dim
                )
                return type(self)(self.stack_shape, a1+a2)
            elif isinstance(other, (int, float, np.integer, np.floating)):
                return type(self)(self.stack_shape, self.array + other)
            else:
                arr, other = self._pad_shape(self.array, other, self.stack_shape)
                return type(self)(self.stack_shape, arr + other)
        def __mul__(self, other):
            if isinstance(other, TensorExpression.ArrayStack):
                a1, a2 = self._make_broadcastable(
                    self.array,
                    other.array,
                    other.stack_dim
                )
                return type(self)(self.stack_shape, a1 * a2)
            elif isinstance(other, (int, float, np.integer, np.floating)):
                return type(self)(self.stack_shape, self.array * other)
            else:
                # print(">>>", self.array.shape, other.shape, self.stack_shape)
                arr, other = self._pad_shape(self.array, other, self.stack_shape)
                # print(arr.shape, other.shape, "<<<")
                return type(self)(self.stack_shape, arr * other)
        def __rmul__(self, other):
            return self*other
        def __pos__(self):
            return self
        def __neg__(self):
            return -1 * self
        def __truediv__(self, other):
            if isinstance(other, TensorExpression.ArrayStack):
                a1, a2 = self._make_broadcastable(
                    self.array,
                    other.array,
                    other.stack_dim
                )
                return type(self)(self.stack_shape, a1 / a2)
            elif isinstance(other, (int, float, np.integer, np.floating)):
                return type(self)(self.stack_shape, self.array / other)
            else:
                arr, other = self._pad_shape(self.array, other, self.stack_shape)
                return type(self)(self.stack_shape, arr / other)
        def __rtruediv__(self, other):
            return other * self.flip()
        def flip(self):
            return type(self)(self.stack_shape, 1/self.array)

        def __pow__(self, power, modulo=None):
            if isinstance(power, TensorExpression.ArrayStack):
                a1, a2 = self._make_broadcastable(
                    self.array,
                    power.array,
                    power.stack_dim
                )
                return type(self)(self.stack_shape, a1 ** a2)
            elif isinstance(power, (int, float, np.integer, np.floating)):
                return type(self)(self.stack_shape, self.array ** power)
            else:
                arr, power = self._pad_shape(self.array, power, self.stack_shape)
                return type(self)(self.stack_shape, arr ** power)

        @property
        def stack_dim(self):
            return len(self.stack_shape)
        @property
        def shape(self):
            return self.array.shape[self.stack_dim:]
        @property
        def ndim(self):
            return self.array.ndim - self.stack_dim
        def expand_dims(self, where):
            if isinstance(where, int):
                where = [where]
            t = self.array
            offset = self.stack_dim
            for i in where:
                if i >= 0:
                    i = offset + i
                t = np.expand_dims(t, i)
            return type(self)(self.stack_shape, t)

        def moveaxis(self, i, j):
            offset = self.stack_dim
            if i >= 0:
                i = offset + i
            if j >= 0:
                j = offset + j
            return type(self)(self.stack_shape, np.moveaxis(self.array, i, j))
        def tensordot(self, other, axes=None):
            if isinstance(other, TensorExpression.ArrayStack):
                other = other.array
            else:
                ogs = other.shape
                for _ in range(self.stack_dim):
                    other = np.expand_dims(other, 0)
                other = np.broadcast_to(other, self.stack_shape + ogs)
            sd = self.stack_dim
            ax1, ax2 = axes
            if isinstance(ax1, int):
                ax1 = [ax1]
            if isinstance(ax2, int):
                ax2 = [ax2]
            axes = [
                [sd + a if a >= 0 else a for a in ax1],
                [sd + b if b >= 0 else b for b in ax2]
            ]
            return type(self)(self.stack_shape, vec_tensordot(self.array, other, shared=len(self.stack_shape), axes=axes))

        def outer(self, other, axes=None):
            if isinstance(other, TensorExpression.ArrayStack):
                other = other.array
            else:
                ogs = other.shape
                for _ in range(self.stack_dim):
                    other = np.expand_dims(other, 0)
                other = np.broadcast_to(other, self.stack_shape + ogs)
            if axes is None:
                exp_axes = [
                    np.arange(self.stack_dim, self.array.ndim),
                    np.arange(self.stack_dim, other.ndim)
                ]
            else:
                sd = self.stack_dim
                exp_axes = [
                    [sd+a if a >= 0 else a for a in axes[0]],
                    [sd+b if b >= 0 else b for b in axes[1]]
                ]
            return type(self)(self.stack_shape, vec_outer(self.array, other, axes=exp_axes))

        def rev_outer(self, other, axes=None):
            if not isinstance(other, TensorExpression.ArrayStack):
                _, other = self._pad_shape(self.array, other, self.stack_shape)
                other = TensorExpression.ArrayStack(self.stack_shape, other)
            return other.outer(self, axes=[axes[1], axes[0]] if axes is not None else axes)

        def __repr__(self):
            return "{}({},\n{}\n)".format(
                type(self).__name__,
                self.stack_shape,
                self.array
            )

    class Term(metaclass=abc.ABCMeta):
        _array_cache = weakref.WeakKeyDictionary()
        # for keeping track of array vals for eval
        def __init__(self, array=None, name=None):
            self._arr = array
            self.name = name
            self._hash = None
        @abc.abstractmethod
        def get_children(self) -> 'Iterable[TensorExpression.Term]':
            raise NotImplementedError("base class")
        @property
        def children(self):
            return self.get_children()
        @abc.abstractmethod
        def deriv(self) -> 'TensorExpression.Term':
            raise NotImplementedError("base class")
        def dQ(self):
            new_term = self.deriv()
            if self.name is not None and isinstance(new_term, TensorExpression.Term):
                new_term.name = 'dQ({})'.format(self.name)
            return new_term
        @abc.abstractmethod
        def array_generator(self, **kwargs) -> np.ndarray:
            raise NotImplementedError("base class")
        @property
        def ndim(self):
            return self.array.ndim
        def get_hash(self):
            return hash((self.name, self.to_string()))
        def __hash__(self):
            if self._hash is None:
                self._hash = self.get_hash()
            return self._hash
        def __eq__(self, other):
            return type(self) is type(other) and hash(self) == hash(other)
        def asarray(self, print_terms=False, cache=True, **kw):
            if self in self._array_cache:  # this allows me to share arrays as long as the hash is right
                return self._array_cache[self]
            else:
                # print(self)
                try:
                    res = self.array_generator(print_terms=print_terms, **kw)
                except TensorExpansionError:
                    raise
                except Exception as e:
                    raise TensorExpansionError('failed to convert {} to an array'.format(self))
                else:
                    if print_terms:
                        print(self)
                        if isinstance(res, TensorExpression.ArrayStack):
                            print(">>>", res.array.shape)
                        else:
                            print(">>>", res.shape)
                    if cache:
                        self._array_cache[self] = res
                        # print(list(self._array_cache.keys()))
                    return res
        @property
        def array(self):
            if self not in self._array_cache: # this allows me to share arrays as long as the hash is right
                if self._arr is None:
                    try:
                        self._arr = self.array_generator()
                    except TensorExpansionError:
                        raise
                    except Exception as e:
                        raise TensorExpansionError('failed to convert {} to an array'.format(self))
                self._array_cache[self] = self._arr
            return self._array_cache[self]
        @array.setter
        def array(self, arr):
            self._arr = arr
        @abc.abstractmethod
        def rank(self) -> int:
            raise NotImplementedError("base class")
        @property
        def ndim(self):
            return self.rank()
        @abc.abstractmethod
        def to_string(self) -> str:
            raise NotImplementedError("base class")
        def __repr__(self):
            if self.name is None:
                name = self.to_string()
                # if name is None or name == "None":
                #     raise ValueError("bad name for type {}".format(type(self).__name__))
                return name
            else:
                return self.name
        @abc.abstractmethod
        def reduce_terms(self, check_arrays=False) -> 'TensorExpression.Term':
            raise NotImplementedError("base class")
        def _check_simp(self, new):
            new._arr = None
            n2 = new.array
            if not np.allclose(self.array, n2):
                raise TensorExpansionError("bad simplification {} -> {}".format(self, new))
        def simplify(self, check_arrays=False):
            if check_arrays:
                arr = self.array
                self._arr = None
                red = self.reduce_terms(check_arrays=True)
                new = red.array
                if not np.allclose(arr, new):
                    raise TensorExpansionError("bad simplification {} -> {}".format(self, red))
                new = red
            else:
                new = self.reduce_terms()
            if isinstance(new, TensorExpression.Term):
                if new.name is None:
                    new.name = self.name
                if new._arr is None:
                    new._arr = self._arr
            return new
        def __add__(self, other):
            return TensorExpression.SumTerm(self, other)
        def __mul__(self, other):
            return TensorExpression.ScalingTerm(self, other)
        def __rmul__(self, other):
            return TensorExpression.ScalingTerm(self, other)

        def __pos__(self):
            return self
        def __neg__(self):
            return -1 * self
        def flip(self):
            raise NotImplementedError("base class")

        def divided(self):
            if self.ndim > 0:
                raise ValueError("term {} isn't a scalar".format(self))
            else:
                return TensorExpression.FlippedTerm(self)
        def __truediv__(self, other):
            if isinstance(other, TensorExpression.Term):
                return TensorExpression.ScalingTerm(self, other.divided())
            else:
                return TensorExpression.ScalingTerm(self, 1 / other)
        def __rtruediv__(self, other):
            if isinstance(other, TensorExpression.Term):
                return other.__truediv__(self)
            else:
                return TensorExpression.ScalingTerm(self.divided(), other)

        def dot(self, other, i, j):
            return TensorExpression.ContractionTerm(self, i, j, other)

        def shift(self, i, j):
            if i < 1 or j < 1 or i == j:
                return self
            else:
                return TensorExpression.AxisShiftTerm(self, i, j)

        def det(self):
            return TensorExpression.DeterminantTerm(self)

        def tr(self, axis1=1, axis2=2):
            return TensorExpression.TraceTerm(self, axis1=axis1, axis2=axis2)

        def inverse(self):
            return TensorExpression.InverseTerm(self)

        def __invert__(self):
            return self.inverse()

        def outer(self, other):
            return TensorExpression.OuterTerm(self, other)

        # def __hash__(self):
        #     return hash(str(self))

        # def __eq__(self, other):
        #     return str(self) == str(other)

    class SumTerm(Term):
        def __init__(self, *terms: 'TensorExpression.Term', array=None, name=None):
            super().__init__(array=array, name=name)
            self.terms = terms
        def get_children(self):
            return self.terms
        def deriv(self):
            return type(self)(*(s.dQ() for s in self.terms))

        def array_generator(self, print_terms=False):
            all_terms = [(i, s.asarray(print_terms=print_terms)) for i, s in enumerate(self.terms)]
            clean_terms = [s for s in all_terms if
                           not (isinstance(s[1], (int, float, np.integer, np.floating)) and s[1] == 0)]
            shp = None
            for s in clean_terms:
                if shp is None:
                    shp = s[1].shape
                if s[1].shape != shp:
                    raise Exception(
                        "this term is bad {} in {} (shape {} instead of shape {})".format(self.terms[s[0]], self,
                                                                                          s[1].shape,
                                                                                          clean_terms[0][1].shape
                                                                                          ))
            clean_terms = [c[1] for c in clean_terms]
            # try:
            if len(clean_terms) > 0:
                s = sum(clean_terms[1:], clean_terms[0])
            else:
                s = 0
            # except:
            #     raise Exception(clean_terms)
            # else:
            return s

        def rank(self):
            return self.terms[0].ndim

        def reduce_terms(self, check_arrays=False):
            full_terms = []
            cls = type(self)
            for t in self.terms:
                t = t.simplify(check_arrays=check_arrays)
                if isinstance(t, cls):
                    full_terms.extend(t.terms)
                elif not (isinstance(t, (int, float, np.integer, np.floating)) and t == 0):
                    full_terms.append(t)
            full_terms = list(sorted(full_terms, key=lambda t: str(t)))  # canonical sorting is string sorting b.c. why not
            return type(self)(*full_terms, array=self._arr)

        def get_hash(self):
            return hash((self.name, tuple(hash(x) for x in self.terms)))
        def to_string(self):
            return '+'.join(str(x) for x in self.terms)

        def substitute(self, other):
            """substitutes other in to the sum by matching up all necessary terms"""
            if isinstance(other, TensorExpression.SumTerm):
                my_terms = list(self.terms)
                for t in other.terms:
                    i = my_terms.index(t)
                    my_terms = my_terms[:i] + my_terms[i + 1:]
                my_terms.append(other)
                return type(self)(*my_terms)
            else:
                my_terms = list(self.terms)
                i = my_terms.index(other)
                my_terms = my_terms[:i] + my_terms[i + 1:]
                my_terms.append(other)
                return type(self)(*my_terms)
    class ScalingTerm(Term):
        def __init__(self, term: 'TensorExpression.Term', scaling, array=None, name=None):
            super().__init__(array=array, name=name)
            self.scaling = scaling
            self.term = term
        def get_children(self):
            return [self.scaling, self.term]
        def rank(self):
            if isinstance(self.scaling, TensorExpression.Term):
                scale_dim = self.scaling.ndim
            elif isinstance(self.scaling, np.ndarray):
                scale_dim = self.scaling.ndim
            else:
                scale_dim = 0
            return scale_dim + self.term.ndim
        def array_generator(self, print_terms=False):
            scaling = self.scaling
            if isinstance(self.scaling, TensorExpression.Term):
                scaling = scaling.asarray(print_terms=print_terms)
            if isinstance(scaling, (np.ndarray, TensorExpression.ArrayStack)): # really doing an outer product...
                term = self.term.asarray(print_terms=print_terms)
                if scaling.ndim > 0 and term.ndim > 0:
                    og_term_dim = term.ndim
                    if isinstance(term, TensorExpression.ArrayStack):
                        term = term.expand_dims([-1]*scaling.ndim)
                    else:
                        for i in range(scaling.ndim):
                            term = np.expand_dims(term, -1)
                    if isinstance(scaling, TensorExpression.ArrayStack):
                        scaling = scaling.expand_dims([0]*og_term_dim)
                    else:
                        for i in range(og_term_dim):
                            scaling = np.expand_dims(scaling, 0)
            else:
                term = self.term.asarray(print_terms=print_terms)
            arr = scaling * term
            return arr

        def to_string(self):
            meh_1 = '({})'.format(self.scaling) if isinstance(self.scaling, TensorExpression.SumTerm) else self.scaling
            meh_2 = '({})'.format(self.term) if isinstance(self.term, TensorExpression.SumTerm) else self.term
            return '{}*{}'.format(meh_1, meh_2)

        def reduce_terms(self, check_arrays=False):
            scaling = self.scaling
            if isinstance(self.scaling, TensorExpression.Term):
                scaling = scaling.simplify(check_arrays=check_arrays)
            if isinstance(scaling, (int, float, np.integer, np.floating)) and scaling == 1:
                return self.term.simplify(check_arrays=check_arrays)
            elif (
                    (isinstance(scaling, (int, float, np.integer, np.floating)) and scaling == 0)
                    or
                    (isinstance(self.term, (int, float, np.integer, np.floating)) and self.term == 0)
            ):
                return 0
            else:
                return type(self)(self.term.simplify(check_arrays=check_arrays), scaling, array=self._arr)

        def deriv(self):
            term = type(self)(self.term.dQ(), self.scaling)
            if isinstance(self.scaling, TensorExpression.Term):
                ugh2 = type(self)(self.term, self.scaling.dQ())
                term += ugh2
            return term

    class ScalarScalingTerm(Term):
        """
        Scaling elementwise with correct broadcasting
        """
        def __init__(self, term: 'TensorExpression.Term', scaling, axes=None, array=None, name=None):
            super().__init__(array=array, name=name)
            self.scaling = scaling
            self.term = term
            self.axes = axes
        def get_children(self):
            return [self.scaling, self.term]
        def rank(self):
            if isinstance(self.scaling, TensorExpression.Term):
                scale_dim = self.scaling.ndim
            elif isinstance(self.scaling, np.ndarray):
                scale_dim = self.scaling.ndim
            else:
                scale_dim = 0
            return scale_dim + self.term.ndim
        def array_generator(self, print_terms=False):
            scaling = self.scaling
            if isinstance(self.scaling, TensorExpression.Term):
                scaling = scaling.asarray(print_terms=print_terms)
            # if isinstance(scaling, (np.ndarray, TensorExpression.ArrayStack)): # really doing an outer product...
            #     term = self.term.asarray(print_terms=print_terms)
            #     if scaling.ndim > 0 and term.ndim > 0:
            #         og_term_dim = term.ndim
            #         if isinstance(term, TensorExpression.ArrayStack):
            #             term = term.expand_dims([-1]*scaling.ndim)
            #         else:
            #             for i in range(scaling.ndim):
            #                 term = np.expand_dims(term, -1)
            #         if isinstance(scaling, TensorExpression.ArrayStack):
            #             scaling = scaling.expand_dims([0]*og_term_dim)
            #         else:
            #             for i in range(og_term_dim):
            #                 scaling = np.expand_dims(scaling, 0)
            # else:

            term = self.term.asarray(print_terms=print_terms)
            # print(self)
            # print(self.axes, term.shape, scaling.shape)
            if (
                    self.axes is None
                    or not isinstance(scaling, (np.ndarray, TensorExpression.ArrayStack))
                    or (
                            self.axes is not None
                            and (
                                    term.ndim - len(self.axes[0]) <= 0 or
                                    scaling.ndim - len(self.axes[1]) <= 0
                            )
                        )
            ):
                # print("???")
                arr = scaling * term
            else: # manage broadcasting for effectively a vecouter
                mask = np.ones(term.ndim, dtype=np.uint8)
                mask[[term.ndim+a if a < 0 else a for a in self.axes[0]]] = 0
                ax1 = np.where(mask)[0]

                mask = np.ones(scaling.ndim, dtype=np.uint8)
                mask[[term.ndim+a if a < 0 else a for a in self.axes[1]]] = 0
                ax2 = np.where(mask)[0]

                axes = [ax1, ax2]
                    # np.setdiff1d(np.arange(term.ndim), [term.ndim+a if a < 0 else a for a in self.axes[0]]),
                    # np.setdiff1d(np.arange(scaling.ndim), [term.ndim+a if a < 0 else a for a in self.axes[1]])
                # ]
                # print(term.shape)
                # print(scaling.shape)
                # print(axes)
                if isinstance(term, TensorExpression.ArrayStack):
                    arr = term.outer(scaling, axes=axes)
                elif isinstance(scaling, TensorExpression.ArrayStack):
                    arr = scaling.rev_outer(term, axes=axes)
                else:
                    arr = vec_outer(term, scaling, axes=axes)
                # print(self)
                # print(arr.shape)
            return arr

        def to_string(self):
            meh_1 = '({})'.format(self.scaling) if isinstance(self.scaling, TensorExpression.SumTerm) else self.scaling
            meh_2 = '({})'.format(self.term) if isinstance(self.term, TensorExpression.SumTerm) else self.term
            return '{}({}*{})'.format(type(self).__name__, meh_2, meh_1)

        def reduce_terms(self, check_arrays=False):
            return self
        #     scaling = self.scaling
        #     if isinstance(self.scaling, TensorExpression.Term):
        #         scaling = scaling.simplify(check_arrays=check_arrays)
        #     if isinstance(scaling, (int, float, np.integer, np.floating)) and scaling == 1:
        #         return self.term.simplify(check_arrays=check_arrays)
        #     elif (
        #             (isinstance(scaling, (int, float, np.integer, np.floating)) and scaling == 0)
        #             or
        #             (isinstance(self.term, (int, float, np.integer, np.floating)) and self.term == 0)
        #     ):
        #         return 0
        #     else:
        #         return type(self)(self.term.simplify(check_arrays=check_arrays), scaling, array=self._arr)

        def deriv(self):
            axes = self.axes
            if axes is None:
                axes = [[0], [0]]
            a = [
                [i+1 if i >= 0 else i for i in axes[0]],
                axes[1]
            ]
            term = type(self)(self.term.dQ(), self.scaling, axes=a)
            if isinstance(self.scaling, TensorExpression.Term):
                b = [
                    axes[1],
                    [i+1 if i >= 0 else i for i in axes[1]]
                ]
                ugh2 = type(self)(self.term, self.scaling.dQ(), axes=b)
                term += ugh2
            return term
    class ScalarPowerTerm(Term):
        """
        Represents x^n.
        Only can get valid derivatives for scalar terms.
        Beware of that.
        """
        def __init__(self, term: 'TensorExpression.Term', pow, array=None, name=None):
            super().__init__(array=array, name=name)
            self.term = term
            self.pow = pow
        def get_children(self):
            return [self.pow, self.term]
        def rank(self):
            return self.term.rank()
        def array_generator(self, print_terms=False):
            wat = self.term.asarray(print_terms=print_terms)
            out = wat ** self.pow
            return out
        def to_string(self):
            return '({}**{})'.format(self.term, self.pow)
        def reduce_terms(self, check_arrays=False):
            if self.pow == 1:
                return self.term.simplify(check_arrays=check_arrays)
            elif self.pow == 0:
                return 1
            elif self.pow == -1:
                return TensorExpression.FlippedTerm(self.term.simplify(check_arrays=True), array=self._arr)
            else:
                return TensorExpression.ScalarPowerTerm(self.term.simplify(check_arrays=check_arrays), self.pow,
                                                        array=self._arr)
        def deriv(self):
            if self.pow == 0:
                return 0 # should probably be doing size stuff but I'm hoping to get away with this...
            elif self.pow == 1:
                return self.term.dQ()
            elif self.pow == 2:
                prod_term = self.term
            else:
                prod_term = TensorExpression.ScalarPowerTerm(self.term, self.pow - 1)
            tdQ = self.term.dQ()
            # print(">>>", tdQ)
            # print(" > ", prod_term)
            # if (
            #         isinstance(prod_term, TensorExpression.Term)
            #         and isinstance(tdQ, TensorExpression.Term)
            #         and tdQ.rank() > 0
            #         and prod_term.rank() > 0
            # ):
            #     return TensorExpression.OuterTerm(
            #         self.pow * tdQ, prod_term
            #     )
            # else:
            return TensorExpression.ScalarScalingTerm(
                tdQ,
                self.pow * prod_term,
                axes=[[-1] if self.term.rank() > 0 else [], [-1]]
            )
    class FlippedTerm(ScalarPowerTerm):
        """
        Represents 1/x. Only can get valid derivatives for scalar terms. Beware of that.
        """

        def __init__(self, term: 'TensorExpression.Term', pow=-1, array=None):
            super().__init__(term, -1, array=array)
        def get_children(self):
            return [self.term]

        def to_string(self):
            return '1/{}'.format(self.term)

        def array_generator(self, print_terms=False):
            out = 1 / self.term.asarray(print_terms=print_terms)
            return out

        def reduce_terms(self, check_arrays=False):
            if isinstance(self.term, TensorExpression.FlippedTerm):
                wat = self.term.term.simplify(check_arrays=check_arrays)
                if check_arrays: self._check_simp(wat)
                return wat
            else:
                return super().reduce_terms(check_arrays=check_arrays)
    class AxisShiftTerm(Term):
        def __init__(self, term: 'TensorExpression.Term', start: int, end: int, array=None, name=None):
            super().__init__(array=array, name=name)
            self.term = term
            self.a = start
            self.b = end
        def get_children(self):
            return [self.term]
        def deriv(self):
            return type(self)(self.term.dQ(), self.a + 1, self.b + 1)

        def array_generator(self, print_terms=False):
            t = self.term.asarray(print_terms=print_terms)
            if isinstance(t, (int, float, np.integer, np.floating)) and t == 0:
                return 0
            else:
                if isinstance(t, TensorExpression.ArrayStack):
                    return t.moveaxis(self.a-1, self.b-1)
                else:
                    return np.moveaxis(t, self.a - 1, self.b - 1)

        def rank(self):
            return self.term.ndim

        def to_string(self):
            return '{{{}#{}->{}}}'.format(self.term, self.a, self.b)

        def reduce_terms(self, check_arrays=False):
            """We simplify over the possible swap classes"""
            cls = type(self)
            simp = self.term.simplify(check_arrays=check_arrays)
            sela = self.a
            selb = self.b
            if sela == selb:
                new = simp
                if check_arrays: self._check_simp(new)
            elif sela > selb:  # we flip this around
                new = cls(simp, selb, sela)
                for i in range(sela - selb - 1):
                    new = cls(new, selb, sela)
                new._arr = self._arr
                new = new.simplify(check_arrays=check_arrays)
                if check_arrays: self._check_simp(new)
            elif isinstance(simp, cls):
                otht = simp.term.simplify(check_arrays=check_arrays)
                otha = simp.a
                othb = simp.b
                if sela == othb:
                    if otha == selb:  # inverses although we shouldn't get here...
                        if otht._arr is None:
                            otht._arr = self._arr
                        new = otht
                    else:  # reduces to a single term
                        new_a = otha
                        new_b = selb
                        new = cls(otht, new_a, new_b, array=self._arr)
                        if check_arrays: self._check_simp(new)
                elif otha > sela:  # we needed to pick a convention...
                    new_sela = sela
                    new_selb = selb if selb < otha else selb + 1
                    other = cls(otht, new_sela, new_selb)
                    new_a = otha if otha > selb else otha - 1
                    new_b = othb if othb > selb else othb - 1
                    new = cls(other, new_a, new_b)
                    if check_arrays: self._check_simp(new)
                    if new._arr is None: new._arr = self._arr
                    return new
                elif otha == sela and othb == selb:
                    # check if we have enough iterates to get the identity
                    is_ident = True
                    for i in range(selb - sela):
                        if isinstance(otht, cls) and otht.a == sela and otht.b == selb:
                            otht = otht.term
                        else:
                            is_ident = False
                            break
                    if is_ident:
                        new = otht
                        if check_arrays:
                            self._check_simp(new)
                    else:
                        new = self
                else:
                    new = cls(simp, sela, selb)
            elif isinstance(simp, TensorExpression.SumTerm):
                new = TensorExpression.SumTerm(
                    *(
                        x.shift(self.a, self.b).simplify(check_arrays=check_arrays)
                        for x in simp.terms
                    ),
                    array=self._arr
                )
                if check_arrays:
                    self._check_simp(new)
            else:
                new = self

            return new

    class OuterTerm(Term):
        """
        Provides an outer product
        """
        def __init__(self,
                     a: 'TensorExpression.Term',
                     b: 'TensorExpression.Term',
                     array=None, name=None
                     ):
            super().__init__(array=array, name=name)
            self.a = a
            self.b = b
        def get_children(self):
            return [self.a, self.b]
        def rank(self):
            return self.a.rank() + self.b.rank()
        def array_generator(self, print_terms=False):
            a = self.a.asarray(print_terms=print_terms)
            b = self.b.asarray(print_terms=print_terms)
            # print(self)
            # print(">", type(a), type(b), a.shape, b.shape)
            og_bdim = b.ndim
            if isinstance(b, TensorExpression.ArrayStack):
                b = b.expand_dims([-1]*a.ndim)
            else:
                for i in range(a.ndim):
                    b = np.expand_dims(b, -1)

            if isinstance(a, TensorExpression.ArrayStack):
                a = a.expand_dims([0]*og_bdim)
            else:
                for i in range(og_bdim):
                    a = np.expand_dims(a, 0)

            return a * b
        def to_string(self):
            meh_1 = '({})'.format(self.a) if isinstance(self.a, TensorExpression.SumTerm) else self.a
            meh_2 = '({})'.format(self.b) if isinstance(self.b, TensorExpression.SumTerm) else self.b
            return '{}<x>{}'.format(meh_1, meh_2)

        def reduce_terms(self, check_arrays=False):
            return self

        # def reduce_terms(self, check_arrays=False):
        #     scaling = self.a
        #     if isinstance(self.scaling, TensorExpression.Term):
        #         scaling = scaling.simplify(check_arrays=check_arrays)
        #     if isinstance(scaling, (int, float, np.integer, np.floating)) and scaling == 1:
        #         return self.term.simplify(check_arrays=check_arrays)
        #     elif (
        #             (isinstance(scaling, (int, float, np.integer, np.floating)) and scaling == 0)
        #             or
        #             (isinstance(self.term, (int, float, np.integer, np.floating)) and self.term == 0)
        #     ):
        #         return 0
        #     else:
        #         return type(self)(self.term.simplify(check_arrays=check_arrays), scaling, array=self._arr)

        def deriv(self):
            return type(self)(self.a.dQ(), self.b) + type(self)(self.a, self.b.dQ())

    class ContractionTerm(Term):
        def __init__(self, left: 'TensorExpression.Term',
                     i: typing.Union[int, typing.Iterable[int]],
                     j: typing.Union[int, typing.Iterable[int]],
                     right: 'TensorExpression.Term', array=None, name=None):
            super().__init__(array=array, name=name)
            self.left = left
            self.i = i
            self.j = j
            self.right = right
        def get_children(self):
            return [self.left, self.right]

        def array_generator(self, print_terms=False):
            t1 = self.left.asarray(print_terms=print_terms)
            t2 = self.right.asarray(print_terms=print_terms)
            if isinstance(t1, (int, float, np.integer, np.floating)) and t1 == 0:
                return 0
            elif isinstance(t2, (int, float, np.integer, np.floating)) and t2 == 0:
                return 0
            else:
                # try:
                i = self.i
                if isinstance(i, (int, np.integer)):
                    i = i - 1
                else:
                    i = [x - 1 for x in i]
                j = self.j
                if isinstance(i, (int, np.integer)):
                    j = j - 1
                else:
                    j = [x - 1 for x in j]
                if isinstance(t1, TensorExpression.ArrayStack):
                    return t1.tensordot(t2, axes=[i, j])
                elif isinstance(t2, TensorExpression.ArrayStack):
                    ogs = t1.shape
                    for _ in range(t2.stack_dim):
                        t1 = np.expand_dims(t1, 0)
                    t1 = np.broadcast_to(t1, t2.stack_shape + ogs)
                    t1 = TensorExpression.ArrayStack(t2.stack_shape, t1)
                    return t1.tensordot(t2, axes=[i, j])
                else:
                    try:
                        return np.tensordot(t1, t2, axes=[i, j])
                    except:
                        raise TensorExpansionError(
                            'failed to contract {}[{}]x{}[{}] for {} and {}'.format(t1.shape, i, t2.shape, j, self.left,
                                                                                    self.right))
                # except (ValueError, IndexError):
                #     raise ValueError("failed to execute {}".format(self))

        def rank(self):
            return self.left.ndim + self.right.ndim - 2

        def deriv(self):
            t1 = type(self)(self.left.dQ(),
                            self.i + 1 if isinstance(self.i, (int, np.integer)) else [x + 1 for x in self.i],
                            # not sure if this is right...
                            self.j,
                            self.right
                            )
            t2 = type(self)(self.left,
                            self.i,
                            self.j + 1 if isinstance(self.j, (int, np.integer)) else [x + 1 for x in self.j],
                            self.right.dQ()
                            )
            return t1 + t2.shift(
                self.left.ndim - (0 if isinstance(self.j, (int, np.integer)) else len(self.j) - 1),
                1
            )

        def to_string(self):
            return '<{}:{},{}:{}>'.format(self.left, self.i, self.j, self.right)

        def reduce_terms(self, check_arrays=False):
            cls = type(self)
            left = self.left.simplify(check_arrays=check_arrays)
            right = self.right.simplify(check_arrays=check_arrays)
            if isinstance(right, TensorExpression.AxisShiftTerm):
                # I flip the names just so it agrees with my math...
                t = right.term
                i = right.a
                j = right.b
                a = self.i
                b = self.j
                n = left.ndim
                if b < i:
                    new = cls(left, a, b, t).shift(n - 2 + i, n - 2 + j)
                    if check_arrays:
                        self._check_simp(new)
                elif b < j:
                    new = cls(left, a, b + 1, t).shift(n - 1 + i, n - 2 + j)
                    if check_arrays:
                        self._check_simp(new)
                elif b == j:
                    new = cls(left, a, i, t)
                    if check_arrays:
                        self._check_simp(new)
                else:
                    new = cls(left, a, b, t).shift(n - 1 + i, n - 1 + j)
                    if check_arrays:
                        self._check_simp(new)

            elif isinstance(left, TensorExpression.AxisShiftTerm):
                # I flip the names just so it agrees with my math...
                t = left.term
                i = left.a
                j = left.b
                a = self.i
                b = self.j
                if a > j:
                    new = cls(t, a, b, right).shift(i, j)
                    if check_arrays:
                        self._check_simp(new)
                elif a == j:
                    new = cls(t, i, b, right)
                    if check_arrays:
                        self._check_simp(new)
                elif a >= i:
                    new = cls(t, a + 1, b, right).shift(i, j - 1)
                    if check_arrays:
                        self._check_simp(new)
                else:
                    new = cls(t, a, b, right).shift(i - 1, j - 1)
                    if check_arrays:
                        self._check_simp(new)
            elif isinstance(left, TensorExpression.SumTerm):
                new = TensorExpression.SumTerm(
                    *(cls(x, self.i, self.j, right).simplify(check_arrays=check_arrays) for x in left.terms),
                    array=self._arr
                )
                if check_arrays:
                    self._check_simp(new)
            elif isinstance(right, TensorExpression.SumTerm):
                new = TensorExpression.SumTerm(
                    *(cls(left, self.i, self.j, x).simplify(check_arrays=check_arrays) for x in right.terms),
                    array=self._arr
                )
                if check_arrays:
                    self._check_simp(new)
            # elif isinstance(self.left, QXTerm) and isinstance(self.right, XVTerm):
            #     new = BasicContractionTerm(self.left.terms, self.left.n, self.i, self.j, self.right.m,
            #                                                      array=self._arr)
            else:
                new = cls(self.left.simplify(check_arrays=check_arrays), self.i, self.j,
                          self.right.simplify(check_arrays=check_arrays), array=self._arr)

            if new._arr is None:
                new._arr = self._arr

            return new

    # fancier terms
    class InverseTerm(Term):
        def __init__(self, term: 'TensorExpression.Term', array=None, name=None):
            super().__init__(array=array, name=name)
            self.term = term

        def get_children(self):
            return [self.term]

        def rank(self):
            return self.term.rank()

        def array_generator(self, print_terms=False):
            t = self.term.asarray(print_terms=print_terms)
            stack_shape = None
            if isinstance(t, TensorExpression.ArrayStack):
                stack_shape = t.stack_shape
                t = t.array
            arr = np.linalg.inv(t)
            if stack_shape is not None:
                arr = TensorExpression.ArrayStack(stack_shape, arr)
            return arr

        def to_string(self):
            return '({}^-1)'.format(self.term)

        def reduce_terms(self, check_arrays=False):
            return type(self)(self.term.simplify(check_arrays=check_arrays), array=self._arr)

        def deriv(self):
            dq = self.term.dQ()
            sub = self.dot(dq, self.ndim, 2).shift(self.ndim, 1)  # self.dot(self.term.dQ(), self.ndim, 2)
            return -sub.dot(self, sub.ndim, 1)  # .shift(self.ndim, 1)
    class TraceTerm(Term):
        def __init__(self, term:'TensorExpression.Term', axis1=1, axis2=2, array=None, name=None):
            super().__init__(array=array, name=name)
            self.term = term
            self.axis1 = axis1
            self.axis2 = axis2

        def get_children(self):
            return [self.term]

        def rank(self):
            return self.term.ndim - 2

        def array_generator(self, print_terms=False):
            t = self.term.asarray(print_terms=print_terms)
            stack_shape = None
            offset = 0
            if isinstance(t, TensorExpression.ArrayStack):
                stack_shape = t.stack_shape
                offset = len(stack_shape)
                t = t.array
            arr = np.trace(t, axis1=offset + self.axis1 - 1, axis2=offset + self.axis2 - 1)
            if stack_shape is not None:
                arr = TensorExpression.ArrayStack(stack_shape, arr)
            return arr

        def to_string(self):
            return 'Tr[{},{}+{}]'.format(self.term, self.axis1, self.axis2)

        def reduce_terms(self, check_arrays=False):
            simp = self.term.simplify(check_arrays=check_arrays)
            if isinstance(simp, TensorExpression.SumTerm):
                new = TensorExpression.SumTerm(
                    *(type(self)(t, axis1=self.axis1, axis2=self.axis2) for t in simp.terms),
                    array=self._arr
                )
                if check_arrays: self._check_simp(new)
            # elif isinstance(simp, AxisShiftTerm):
            #     # wait this is completely wrong...?
            #
            #     sub = type(self)(simp.term, axis1=self.axis1, axis2=self.axis2)
            #     new = AxisShiftTerm(
            #         type(self)(simp.term, axis1=self.axis1, axis2=self.axis2),
            #         simp.a,
            #         simp.b,
            #         array=self._arr
            #     )
            #     if check_arrays: self._check_simp(new)
            else:
                new = type(self)(simp, axis1=self.axis1, axis2=self.axis2, array=self._arr)
            return new

        def deriv(self):
            wat = self.term.dQ()
            return type(self)(wat, axis1=self.axis1 + 1, axis2=self.axis2 + 1)
    class DeterminantTerm(Term):
        def __init__(self, term: 'TensorExpression.Term', array=None, name=None):
            super().__init__(array=array, name=name)
            self.term = term
        def get_children(self):
            return [self.term]
        def rank(self):
            return 0
        def array_generator(self, print_terms=False):
            t = self.term.asarray(print_terms=print_terms)
            stack_shape = None
            if isinstance(t, TensorExpression.ArrayStack):
                stack_shape = t.stack_shape
                t = t.array
            arr = np.linalg.det(t)
            if stack_shape is not None:
                arr = TensorExpression.ArrayStack(stack_shape, arr)
            return arr
        def to_string(self):
            return '|{}|'.format(self.term)
        def reduce_terms(self, check_arrays=False):
            return type(self)(self.term.simplify(check_arrays=check_arrays), array=self._arr)
        def deriv(self):
            inv_dot = TensorExpression.InverseTerm(self.term).dot(
                self.term.dQ(),
                self.term.ndim,
                2
            )
            if self.term.ndim > 1:
                inv_dot = inv_dot.shift(self.term.ndim, 1)
            # print(inv_dot)
            tr = TensorExpression.TraceTerm(
                inv_dot,
                axis1=2,
                axis2=inv_dot.ndim
            ) * self
            # if self.term.ndim > 2:
            #     tr = tr.shift(self.term.ndim-1, 1)
            return tr
    class VectorNormTerm(Term):
        def __init__(self, term: 'TensorExpression.Term', array=None, name=None, axis=-1):
            super().__init__(array=array, name=name)
            self.term = term
            self.axis = axis
        def get_children(self):
            return [self.term]
        def rank(self):
            return self.term.rank() - 1
        def array_generator(self, print_terms=False):
            t = self.term.asarray(print_terms=print_terms)
            stack_shape = None
            offset = 0
            if isinstance(t, TensorExpression.ArrayStack):
                stack_shape = t.stack_shape
                offset = len(stack_shape)
                t = t.array
            ax = (self.axis + offset) if self.axis >= 0 else self.axis
            arr = np.linalg.norm(t, axis=ax)
            # arr = np.ones(arr.shape) if isinstance(arr, np.ndarray) else arr
            if stack_shape is not None:
                arr = TensorExpression.ArrayStack(stack_shape, arr)
            return arr
        def to_string(self):
            return '||{}||'.format(self.term)
        def reduce_terms(self, check_arrays=False):
            return type(self)(self.term.simplify(check_arrays=check_arrays), array=self._arr)
        def deriv(self):
            return self.term / self

    class ScalarFunctionTerm(Term):
        def __init__(self, term, name='f', f=None, array=None, derivative_order=0):
            super().__init__(array=array)
            self.term = term
            self.fname = name
            if f is not None and not isinstance(f, dict):
                f = {'function':f, 'derivatives':None}
            self.func = f
            self.deriv_order = derivative_order
        def get_children(self):
            return [self.term]
        def rank(self):
            return self.term.rank() + self.deriv_order
        def array_generator(self, print_terms=False):
            t = self.term.asarray(print_terms=print_terms)
            stack_shape = None
            if isinstance(t, TensorExpression.ArrayStack):
                stack_shape = t.stack_shape
                t = t.array
            if self.func is None:
                raise NotImplementedError("need concrete function to evaluate")
            if self.deriv_order == 0:
                f = self.func['function']
                if f is None:
                    raise NotImplementedError("need concrete function to evaluate")
                res = f(t)
            else:
                df = self.func['derivatives']
                if df is None:
                    raise NotImplementedError("need concrete derivatives to evaluate")
                res = df(self.deriv_order)(t)
            if stack_shape is not None:
                res = TensorExpression.ArrayStack(stack_shape, res)
            return res
        def to_string(self):
            return (
                "{}({})".format(self.fname, self.term)
                if self.deriv_order == 0 else
                "{}[{}']({})".format(self.fname, self.deriv_order, self.term)
            )
        def reduce_terms(self, check_arrays=False):
            return type(self)(self.term.simplify(check_arrays=check_arrays),
                              name=self.fname, f=self.func, array=self._arr,
                              derivative_order=self.deriv_order
                              )
        def deriv(self):
            return TensorExpression.ScalarScalingTerm(
                self.term.deriv(),
                type(self)(
                    self.term, name=self.fname, f=self.func, array=None,
                    derivative_order=self.deriv_order + 1
                ),
                axes=[[-1] if self.term.rank() > 0 else [], [-1] if self.term.rank() > 0 else []]
                #TODO: I _know_ this is wrong in general
                #      but I'm not sure about all the correct rules yet
            )
    class ConstantArray(Term):
        """
        Square tensor of constants (squareness assumed, not checked)
        """
        def __init__(self, array, parent:'TensorExpression.Term'=None, name=None):
            self.base_array = np.asanyarray(array)
            super().__init__(array=None, name=name)
            self.parent = parent
        def get_children(self): # really what this term needs to be evaluated before it can be evaluated...
            return [self.parent]
        def rank(self):
            return self.base_array.ndim
        def array_generator(self, print_terms=False):
            a = self.base_array
            if not isinstance(a, TensorExpression.ArrayStack) and self.parent is not None:
                p = self.parent.asarray()
                if isinstance(p, TensorExpression.ArrayStack):
                    ogs = a.shape
                    for _ in p.stack_shape:
                        a = np.expand_dims(a, 0)
                    a = np.broadcast_to(a, p.stack_shape + ogs)
                    a = TensorExpression.ArrayStack(p.stack_shape, a)
            return a
        def get_hash(self):
            if isinstance(self._arr, np.ndarray):
                buf = self._arr.view(np.uint8)
                h = hashlib.sha1(buf).hexdigest()
            elif self._arr is None:
                if self.base_array is not None:
                    buf = self.base_array.view(np.uint8)
                    h = hashlib.sha1(buf).hexdigest()
                else:
                    h = hash(self.base_array)
            else:
                h = hash(self._arr)
            return hash((self.name, h))
        def to_string(self):
            if self._arr is None:
                return str(self.base_array)
            else:
                return str(self._arr)
        def deriv(self):
            a = self.base_array
            return TensorExpression.ConstantArray(np.zeros(a.shape + (a.shape[-1],)), parent=self)
        def reduce_terms(self, check_arrays=False):
            return self
    class IdentityMatrix(ConstantArray):
        def __init__(self, ndim, parent=None, name="I"):
            super().__init__(np.eye(ndim), parent=parent, name=name)

    class OuterPowerTerm(Term):
        """
        Represents a matrix-power type term
        """
        def __init__(self, base:'TensorExpression.Term', pow:int, array=None, name=None):
            if pow<1:
                raise NotImplementedError("???")
            super().__init__(array=array, name=name)
            self.base = base
            self.pow = pow
        def get_children(self):
            return [self.base]
        def rank(self):
            return self.base.rank() * self.pow
        def array_generator(self, print_terms=False):
            # print(hash(self))
            # allows caching to be reused hopefully...
            if self.pow == 1:
                return self.base.asarray(print_terms=print_terms)
            else:
                return type(self)(self.base, self.pow-1).outer(self.base).asarray(print_terms=print_terms)
        def get_hash(self):
            if self.pow == 1:
                return hash(self.base)
            else:
                return hash((self.name, hash(self.base), self.pow))
        def to_string(self):
            return "{}({}, {})".format(type(self).__name__, self.base, self.pow)
        def deriv(self):
            if self.pow > 2:
                return TensorExpression.OuterTerm(type(self)(self.base, self.pow-1), self.base).dQ()
            elif self.pow == 2:
                return TensorExpression.OuterTerm(self.base, self.base).dQ()
            else:
                return self.base.dQ()
        def reduce_terms(self, check_arrays=False):
            return self
    class TermVector(Term):
        def __init__(self, *terms:'TensorExpression.Term|List[TensorExpression.Term]', array=None, name=None):
            if len(terms) == 1 and not isinstance(terms[0], TensorExpression.Term):
                terms = terms[0]
            self.terms = terms
            super().__init__(array=array, name=name)
        def get_children(self):
            return self.terms
        def rank(self):
            r = self.terms[0].rank()
            if all(t.rank() == r for t in self.terms):
                return r+1
            else:
                return 1
        def array_generator(self, print_terms=False):
            return np.array([
                t.asarray(print_terms=print_terms) for t in self.terms
            ])
        def to_string(self):
            return str(self.terms)
        def deriv(self):
            return type(self)([
                t.deriv() for t in self.terms
            ])
        def reduce_terms(self, check_arrays=False):
            return self
    class CoordinateVector(Term):
        def __init__(self, vals_array, array=None, name=None):
            if not isinstance(vals_array, int) and array is None:
                if isinstance(vals_array, np.ndarray) or isinstance(vals_array[0], (int, float, np.integer, np.floating)):
                    vals_array = np.asanyarray(vals_array)
                    array = vals_array
            super().__init__(array=array, name=name)
            self.base_array = vals_array
        def get_children(self):
            if isinstance(self.base_array, (int, np.integer)):
                return []
            elif not isinstance(self.base_array, (np.ndarray, TensorExpression.ArrayStack)):
                return list(self.base_array)
            else:
                return []
        def rank(self):
            return 1
        def array_generator(self, print_terms=False):
            if isinstance(self.base_array, (np.ndarray, TensorExpression.ArrayStack)):
                return self.base_array
            else:
                raise Exception("need explicit array")
        def to_string(self):
            return "{}({})".format(type(self).__name__, self.base_array)
        def deriv(self):
            if isinstance(self.base_array, int):
                return TensorExpression.IdentityMatrix(self.base_array, parent=self)
            else:
                return TensorExpression.IdentityMatrix(len(self.base_array), parent=self)
        def reduce_terms(self, check_arrays=False):
            return self
    class CoordinateTerm(Term):
        def __init__(self,
                     idx:int,
                     vec:'TensorExpression.CoordinateVector',
                     array=None, name=None
                     ):
            super().__init__(array=array, name=name)
            self.vec = vec
            self.idx = idx
        def get_children(self):
            return []
        def rank(self):
            return 0
        def array_generator(self, print_terms=False):
            return self.vec.asarray(print_terms=print_terms)[self.idx]
        def to_string(self):
            return "{}({}, {})".format(type(self).__name__, self.vec, self.idx)
        def deriv(self):
            vec = np.zeros(len(self.vec.base_array))
            vec[self.idx] = 1
            return TensorExpression.ConstantArray(vec)
        def reduce_terms(self, check_arrays=False):
            return self
    class PolynomialTerm(Term):
        def __init__(self,
                     expansion:'Taylor.FunctionExpansion',
                     coords:'TensorExpression.TermVector'=None,
                     array=None, name=None
                     ):
            super().__init__(array=array, name=name)
            self.expansion = expansion
            self.coords = coords
        def get_children(self):
            return [self.coords]
        def rank(self):
            return 0
        def array_generator(self, print_terms=False):
            arr = self.coords.asarray(print_terms=print_terms)
            stack_shape = None
            if isinstance(arr, TensorExpression.ArrayStack):
                stack_shape = arr.stack_shape
                arr = arr.array
            res = self.expansion(arr)
            if stack_shape is not None:
                res = TensorExpression.ArrayStack(stack_shape, res)
            return res
        def to_string(self):
            return "{}({})".format(self.expansion, self.coords)
        def deriv(self):
            return type(self)(self.expansion.deriv(), coords=self.coords)
        def reduce_terms(self, check_arrays=False):
            return self

class TensorExpansionError(Exception):
    ...
class TensorExpansionTerms:
    """
    A friend of DumbTensor which exists
    to not only make the tensor algebra suck less but also
    to make it automated by making use of some simple rules
    for expressing derivatives specifically in the context of
    doing the coordinate transformations we need to do.
    Everything here is 1 indexed since that's how I did the OG math
    """
    def __init__(self, qx_terms, xv_terms, qxv_terms=None, base_qx=None, base_xv=None,
                 q_name='Q',
                 v_name='V'
                 ):
        """
        :param qx_terms:
        :type qx_terms: Iterable[np.ndarray]
        :param xv_terms:
        :type xv_terms: Iterable[np.ndarray]
        """
        self.qx_terms = qx_terms
        self.xv_terms = [0.]*len(qx_terms) if xv_terms is None else xv_terms
        self.qxv_terms = qxv_terms
        self.base_qx = base_qx
        self.base_xv = base_xv
        self.q_name = q_name
        self.v_name = v_name
    def QX(self, n):
        return self.QXTerm(self, n)
    def XV(self, m):
        return self.XVTerm(self, m)
    def QXV(self, n, m):
        return self.QXVTerm(self,n,  m)

    class QXTerm(TensorExpression.Term):
        def __init__(self, terms: 'TensorExpansionTerms', n: int, array=None):
            super().__init__(array=array)
            self.terms = terms
            self.n = n
        def get_children(self):
            return []
        def deriv(self):
            return type(self)(self.terms, self.n + 1)
        def array_generator(self, print_terms=False):
            if self.n == 0:
                return self.terms.base_qx
            else:
                return self.terms.qx_terms[self.n - 1]
        def rank(self):
            if isinstance(self.array, (int, float, np.integer, np.floating)) and self.array == 0:
                return self.n + 1
            else:
                return self.array.ndim

        def to_string(self):
            return '{}[{}]'.format(self.terms.q_name, self.n)

        def reduce_terms(self, check_arrays=False):
            return self
    class XVTerm(TensorExpression.Term):
        def __init__(self, terms: 'TensorExpansionTerms', m: int, array=None):
            super().__init__(array=array)
            self.terms = terms
            self.m = m
        def get_children(self):
            return []
        def deriv(self):
            mixed_terms = self.terms.qxv_terms
            if (
                    mixed_terms is not None
                    and self.m > 0
                    and len(mixed_terms) > 0
                    # and mixed_terms[0] is not None
                    and len(mixed_terms[0]) >= self.m
                    and mixed_terms[0][self.m - 1] is not None
            ):
                return self.terms.QXV(1, self.m)
            else:
                return TensorExpression.ContractionTerm(self.terms.QX(1), 2, 1, self.terms.XV(self.m + 1))
        def array_generator(self, print_terms=False):
            if self.m == 0:
                return self.terms.base_qx
            else:
                return self.terms.xv_terms[self.m - 1]
        def rank(self):
            if isinstance(self.array, (int, float, np.integer, np.floating)) and self.array == 0:
                return self.m
            else:
                return self.array.ndim
        def to_string(self):
            return '{}[{}]'.format(self.terms.v_name, self.m)
        def reduce_terms(self, check_arrays=False):
            return self
    class QXVTerm(TensorExpression.Term):
        def __init__(self, terms: 'TensorExpansionTerms', n: int, m: int, array=None):
            super().__init__(array=array)
            self.terms = terms
            self.n = n
            self.m = m
        def get_children(self):
            return []
        def deriv(self):
            return type(self)(self.terms, self.n + 1, self.m)

        def array_generator(self, print_terms=False):
            return self.terms.qxv_terms[self.n - 1][self.m - 1]

        def rank(self):
            if isinstance(self.array, (int, float, np.integer, np.floating)) and self.array == 0:
                return self.n + self.m
            else:
                return self.array.ndim

        def to_string(self):
            return '{}{}[{},{}]'.format(self.terms.q_name, self.terms.v_name, self.n, self.m)

        def reduce_terms(self, check_arrays=False):
            return self
    class BasicContractionTerm(TensorExpression.Term):
        """
        Special case tensor contraction term between two elements of the
        tensor expansion terms.
        """

        def __init__(self, terms: 'TensorExpansionTerms', n: int, i: int, j: int, m: int, array=None):
            super().__init__(array=array)
            self.terms = terms
            self.n = n
            self.m = m
            self.i = i
            self.j = j

        def deriv(self):
            return TensorExpression.SumTerm(
                type(self)(
                    self.terms,
                    self.n + 1,
                    self.i + 1,
                    self.j,
                    self.m
                ),
                TensorExpression.ContractionTerm(
                    self.terms.QX(1),
                    2,
                    self.n,
                    type(self)(
                        self.terms,
                        self.n,
                        self.i,
                        self.j + 1,
                        self.m + 1
                    )
                )
            )

        def array_generator(self, print_terms=False):
            arr = np.tensordot(self.terms.qx_terms[self.n - 1], self.terms.xv_terms[self.m - 1], axes=[self.i, self.j])
            return arr

        def rank(self):
            return self.terms.qx_terms[self.n - 1].ndim + self.terms.xv_terms[self.m - 1].ndim - 2

        def to_string(self):
            return '<Q[{}]:{},{}:V[{}]>'.format(self.n, self.i, self.j, self.m)

        def reduce_terms(self, check_arrays=False):
            return self

class TensorDerivativeConverter:
    """
    A class that makes it possible to convert expressions
    involving derivatives in one coordinate system in another
    """

    TensorExpansionError=TensorExpansionError

    #TODO: add way to not recompute terms over and over
    def __init__(self,
                 jacobians,
                 derivatives=None,
                 mixed_terms=None,
                 jacobians_name='Q',
                 values_name='V'
                 ):
        """

        :param jacobians: The Jacobian and higher-order derivatives between the coordinate systems
        :type jacobians: Iterable[np.ndarray]
        :param derivatives: Derivatives of some quantity in the original coordinate system
        :type derivatives: Iterable[np.ndarray]
        :param mixed_terms: Mixed derivatives of some quantity involving the new and old coordinates
        :type mixed_terms: Iterable[Iterable[None | np.ndarray]]
        """

        if derivatives is None:
            derivatives = [0] * len(jacobians)

        if jacobians[0].ndim > 2:
            stack_shape = jacobians[0].shape[:-2]
            jacobians = [
                TensorExpression.ArrayStack(
                    stack_shape,
                    j
                )
                for j in jacobians
            ]
            derivatives = [
                TensorExpression.ArrayStack(
                    stack_shape,
                    d
                ) if not (isinstance(d, (int, float, np.integer, np.floating)) and d == 0) else 0
                for d in derivatives
            ]

        self.terms = TensorExpansionTerms(jacobians, derivatives, qxv_terms=mixed_terms, q_name=jacobians_name, v_name=values_name)

    def convert(self, order=None, print_transformations=False, check_arrays=False):

        if order is None:
            order = len(self.terms.qx_terms)

        if order < 0:
            raise ValueError("cannot convert derivatives between coordinates systems to order 0")

        arrays = []
        deriv = self.terms.QX(1).dot(self.terms.XV(1), 2, 1)
        arrays.append(deriv.array)

        for i in range(2, order+1):
            # print(">>>>>", deriv)
            deriv = deriv.dQ().simplify(check_arrays=check_arrays) # there's an occasional issue with shift simplifications I think...
            if print_transformations:
                print(">> order: ", i, deriv)
            arrays.append(deriv.array)

        arrays = [
            a.array if isinstance(a, TensorExpression.ArrayStack) else a
            for a in arrays
        ]
        return arrays

    @classmethod
    def compute_partition_terms(cls, partition):
        _, counts = np.unique(partition, return_counts=True)

        # compute the reduced multinomial coefficient in stable (enough) form
        multinomial_num = np.flip(np.arange(1, 1 + np.sum(partition)))
        multinomial_denum = np.concatenate([
            np.arange(1, 1 + j)
            for j in partition
        ])
        multi_redux_terms = np.concatenate([
            np.arange(1, 1 + j)
            for j in counts
        ])
        full_denom = np.flip(np.sort(np.concatenate([multinomial_denum, multi_redux_terms])))
        numl = len(multinomial_num)
        denl = len(full_denom)
        if numl > denl:
            full_denom = np.concatenate([full_denom, np.ones(numl - denl, dtype=full_denom.dtype)])
        elif denl > numl:
            multinomial_num = np.concatenate([multinomial_num, np.ones(denl - numl, dtype=multinomial_num.dtype)])

        return np.prod(multinomial_num / full_denom)

    @classmethod
    def convert_partition(cls, partition, derivs, vals, val_axis=0):
        if len(vals) <= len(partition) - 1:
            return 0
        base_term = vals[len(partition) - 1]
        if isinstance(base_term, (int, float, np.integer, np.floating)) and base_term == 0:
            return 0
        perm_counter = len(partition)
        perm_idx = []  # to establish the set of necessary permutations to make things symmetric
        if val_axis < 0:
            val_axis = base_term.ndim + val_axis
        for i in partition:
            if len(derivs) <= i - 1:
                return 0

            d = derivs[i - 1]
            if isinstance(d, (int, float, np.integer, np.floating)) and d == 0:
                return 0
            base_term = np.tensordot(d, base_term, axes=[-1, val_axis])
            if i > 1:
                perm_idx.extend([perm_counter] * i)
                perm_counter -= 1
            else:
                perm_idx.append(perm_counter)
            val_axis = min(base_term.ndim - 1, val_axis + d.ndim - 1)

        # sometimes we overcount, so we factor that out here
        nterms = cls.compute_partition_terms(partition)
        perm_inds, _ = UniquePermutations(perm_idx).permutations(return_indices=True)

        overcount = len(perm_inds) / nterms
        base_term = base_term / overcount

        return sum(
            base_term.transpose(list(p) + list(range(len(p), base_term.ndim)))
            for p in perm_inds
        )

    @classmethod
    def convert_fast(cls, derivs, vals, val_axis=-1, order=None):
        terms = []
        if order is None:
            order = len(vals)

        for o in range(1, order+1):
            term = sum(
                cls.convert_partition(p, derivs, vals, val_axis=val_axis)
                for parts in IntegerPartitioner.partitions(o)
                for p in parts
            )
            terms.append(term)

        return terms



import os.path

import numpy as np
from collections import namedtuple
from ..Parsers import ElectronicStructureLogReader
from ...Parsers import FileStreamReader, XYZParser, Word, Number


__all__ = [
    "CRESTParser"
]

class CRESTOptLogParser(FileStreamReader):

    def __init__(self, file, **kwargs):
        super().__init__(file, **kwargs)

    CRESTCoords = namedtuple("CRESTCoords", ["energy", "atoms", "coords"])
    @classmethod
    def parse_struct(cls, data):
        coords = np.array(Number.findall(data)).astype(float)
        e, coords = coords[0], coords[1:].reshape(-1, 3)
        atoms = Word.findall(data)
        return cls.CRESTCoords(
            e,
            atoms,
            coords
        )

    def get_next_block(self):
        block = self.get_tagged_block("=", "Etot")
        if block is None:
            return None
        else:
            return block.rsplit("\n", 1)[0]

    def parse(self):
        res = []
        block = self.get_next_block()
        while block is not None:
            res.append(self.parse_struct(block))
            block = self.get_next_block()

        return res

class CRESTConfgenLogParser(ElectronicStructureLogReader):
    components_name = "CRESTLogComponents"

    def parse(self, keys=None, num=None, reset=False):
        if keys is None:
            keys = [
                "CommandLine",
                "CalculationInfo",
                "InputStructure",
                "FinalOptInfo",
                "FinalEnsembleInfo"
            ]
        return super().parse(keys, num=num, reset=reset)

class CRESTParser:

    opt_log_file = 'crestopt.log'
    confgen_log_file = 'confgen.log'
    ensemble_energies_file = 'ensemble_energies.log'
    ensembe_file = 'ensemble_energies.log'
    conformers_best_file = 'crest_best.xyz'
    conformers_file = 'crest_conformers.xyz'
    rotamers_file = 'crest_rotamers.xyz'
    def __init__(self, parse_dir,
                 opt_log_file=None,
                 confgen_log_file=None,
                 ensemble_energies_file=None,
                 conformers_file=None,
                 conformers_best_file=None,
                 rotamers_file=None
                 ):
        self.dir = parse_dir
        self.opt_log_file = self._locate_file(opt_log_file, self.opt_log_file)
        self.confgen_log_file = self._locate_file(confgen_log_file, self.confgen_log_file)
        self.ensemble_energies_file = self._locate_file(ensemble_energies_file, self.ensemble_energies_file)
        self.conformers_file = self._locate_file(conformers_file, self.conformers_file)
        self.conformers_best_file = self._locate_file(conformers_best_file, self.conformers_best_file)
        self.rotamers_file = self._locate_file(rotamers_file, self.rotamers_file)
    def parse_optimized_structures(self):
        with CRESTOptLogParser(self.opt_log_file) as parser:
            structs = parser.parse()
        return structs

    def parse_ensemble_enegies(self):
        return np.loadtxt(self.ensemble_energies_file)

    CRESTConformers = namedtuple("CRESTConformers", ['atoms', 'energies', 'coords'])
    def parse_conformers(self, conformers_file=None):
        if conformers_file is None:
            conformers_file = self.conformers_file
        with XYZParser(conformers_file) as parser:
            structs = parser.parse()
        atoms = structs[0][1]
        coords = np.array([s for c,a,s in structs])
        energies = np.array([c.strip() for c,a,s in structs]).astype(float)
        return self.CRESTConformers(atoms, energies, coords)

    def parse_best_conformers(self):
        return self.parse_conformers(self.conformers_best_file)

    CRESTRotamers = namedtuple("CRESTRotamers", ['atoms', 'energies', 'weights', 'coords'])
    def parse_rotamers(self, rotamers_file=None):
        if rotamers_file is None:
            rotamers_file = self.rotamers_file
        with XYZParser(rotamers_file) as parser:
            structs = parser.parse()
        atoms = structs[0][1]
        coords = np.array([s for c,a,s in structs])
        eng_rel = np.array([c.split()[:2] for c,a,s in structs]).astype(float)
        return self.CRESTRotamers(atoms, eng_rel[:, 0], eng_rel[:, 1], coords)

    def parse_log(self):
        with CRESTConfgenLogParser(self.confgen_log_file) as parser:
            structs = parser.parse()
        return structs

    def _locate_file(self, base_file, default_file):
        if base_file is None:
            base_file = default_file
        test = os.path.join(self.dir, base_file)
        if os.path.exists(test):
            return test
        else:
            return base_file



import importlib
from ...Parsers import FileStreamReader, FileStreamCheckPoint, FileStreamReaderException

__all__ = [
    "ElectronicStructureLogReader"
]

class ElectronicStructureLogReader(FileStreamReader):
    """
    Implements a stream based reader for a generic electronic structure .log file.
    This is inherits from the `FileStreamReader` base, and takes a two pronged approach to getting data.
    First, a block is found in a log file based on a pair of tags.
    Next, a function (usually based on a `StringParser`) is applied to this data to convert it into a usable data format.
    The goal is to move toward wrapping all returned data in a `QuantityArray` so as to include data type information, too.
    """

    components_name = None
    components_package = ".LogComponents"
    _comps = None
    @classmethod
    def load_components(cls):
        if cls._comps is None:
            pkg_root = cls.components_package
            if pkg_root.startswith('.'):
                pkg_root = cls.__module__.rsplit('.', 1)[0] + cls.components_package
            cls._comps = importlib.import_module(pkg_root + '.' + cls.components_name)
        return cls._comps
    @property
    def registered_components(self):
        return self.load_components().__components__
    @property
    def default_keys(self):
        return self.load_components().__defaults__
    @property
    def default_ordering(self):
        return self.load_components().__ordering__

    def parse(self, keys, num=None, reset=False):
        """The main function we'll actually use. Parses bits out of a .log file.

        :param keys: the keys we'd like to read from the log file
        :type keys: str or list(str)
        :param num: for keys with multiple entries, the number of entries to pull
        :type num: int or None
        :return: the data pulled from the log file, strung together as a `dict` and keyed by the _keys_
        :rtype: dict
        """
        # if keys is None:
        #     keys = self.get_default_keys()
        # important for ensuring correctness of what we pull
        if isinstance(keys, str):
            keys = (keys,)
        keys = sorted(keys,
                      key = lambda k: (
                          -1 if (self.registered_components[k]["mode"] == "List") else (
                              self.default_ordering[k] if k in self.default_ordering else 0
                          )
                      )
                      )

        res = {}
        if reset:
            with FileStreamCheckPoint(self):
                for k in keys:
                    comp = self.registered_components[k]
                    res[k] = self.parse_key_block(**comp, num=num)
        else:
            for k in keys:
                comp = self.registered_components[k]
                try:
                    res[k] = self.parse_key_block(**comp, num=num)
                except:
                    raise FileStreamReaderException("failed to parse block for key '{}'".format(k))
        return res

    # job_default_keys = {
    #     "opt":{
    #         "p": ("StandardCartesianCoordinates", "OptimizedScanEnergies", "OptimizedDipoleMoments"),
    #         "_": ("StandardCartesianCoordinates", "OptimizedScanEnergies")
    #     },
    #     "popt": {
    #         "p": ("StandardCartesianCoordinates", "OptimizedScanEnergies", "OptimizedDipoleMoments"),
    #         "_": ("StandardCartesianCoordinates", "OptimizedScanEnergies")
    #     },
    #     "scan": ("StandardCartesianCoordinates", "ScanEnergies")
    # }
    # def get_default_keys(self):
    #     """
    #     Tries to get the default keys one might be expected to want depending on the type of job as determined from the Header
    #     Currently only supports 'opt', 'scan', and 'popt' as job types.
    #
    #     :return: key listing
    #     :rtype: tuple(str)
    #     """
    #     header = self.parse("Header", reset=True)["Header"]
    #
    #     header_low = {k.lower() for k in header.job}
    #     for k in self.job_default_keys:
    #         if k in header_low:
    #             sub = self.job_default_keys[k]
    #             if isinstance(sub, dict):
    #                 for k in sub:
    #                     if k in header_low:
    #                         defs = sub[k]
    #                         break
    #                 else:
    #                     defs = sub["_"]
    #             else:
    #                 defs = sub
    #             break
    #     else:
    #         raise FileStreamReaderException("unclear what default keys should be used if not a scan and not a popt")
    #
    #     return ("Header", ) + tuple(defs) + ("Footer",)

    @classmethod
    def read_props(cls, file, keys):
        with cls(file) as reader:
            parse = reader.parse(keys)
        if isinstance(keys, str):
            parse = parse[keys]
        return parse